"""Grading function for ScienceBench task 15 (admet_ai)."""

from __future__ import annotations

import pandas as pd
from sklearn.metrics import roc_auc_score

REQUIRED_COLUMNS = {"Drug", "Y"}
AUC_THRESHOLD = 0.84


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    """Return 1.0 if ordering matches and ROC AUC >= threshold, else 0.0."""
    if submission.empty:
        print("Submission is empty.")
        return 0.0

    if not REQUIRED_COLUMNS.issubset(submission.columns):
        print(f"Submission missing required columns: {REQUIRED_COLUMNS - set(submission.columns)}")
        return 0.0

    if not REQUIRED_COLUMNS.issubset(answers.columns):
        print("Answer file is missing required columns.")
        return 0.0

    if list(submission["Drug"]) != list(answers["Drug"]):
        print("Drug ordering mismatch.")
        return 0.0

    try:
        auc = roc_auc_score(answers["Y"].values, submission["Y"].values)
    except ValueError as exc:
        print(f"Unable to compute ROC AUC: {exc}")
        return 0.0

    print(f"ROC AUC: {auc}")
    return 1.0 if auc >= AUC_THRESHOLD else 0.0



import re
import shutil
from pathlib import Path
from typing import List

import pandas as pd
from tqdm import tqdm


def _create_split(
    train_files: List[Path],
    test_files: List[Path],
    public_path: Path,
    private_path: Path,
):
    """
    Helper function to process and save a single train/test split.

    This function handles file copying, renaming, zipping, and the creation of
    answer and sample submission files for a given set of train/test files.
    """
    # Create output directories if they don't exist
    public_path.mkdir(exist_ok=True, parents=True)
    private_path.mkdir(exist_ok=True, parents=True)

    # Process and copy train files
    train_output_dir = public_path / "train2"
    train_output_dir.mkdir(exist_ok=True)
    # Sort files to ensure deterministic indexing
    sorted_train_files = sorted(train_files)
    for idx, sample in enumerate(
        tqdm(sorted_train_files, desc=f"Creating train set for {public_path.name}")
    ):
        new_sample_name = re.sub(r"TRAIN\d+", f"TRAIN{idx}", sample.name)
        new_sample = train_output_dir / new_sample_name
        shutil.copy(sample, new_sample)

    # Process and copy test files, collecting answers
    answer_rows = []
    test_output_dir = public_path / "test2"
    test_output_dir.mkdir(exist_ok=True)
    # Sort files to ensure deterministic indexing
    sorted_test_files = sorted(test_files)
    for idx, sample in enumerate(
        tqdm(sorted_test_files, desc=f"Creating test set for {public_path.name}")
    ):
        new_sample_name = sample.name.split("TRAIN")[0] + f"Test{idx}.aif"
        new_sample = test_output_dir / new_sample_name
        shutil.copy(sample, new_sample)
        answer_rows.append(
            {"clip": new_sample_name, "probability": 1 if sample.stem.endswith("_1") else 0}
        )

    # Assertions to verify file counts
    assert len(sorted_train_files) == len(list(train_output_dir.glob("*.aif")))
    assert len(sorted_test_files) == len(list(test_output_dir.glob("*.aif")))

    # Create zipped versions and remove temporary unzipped directories
    shutil.make_archive(public_path / "train2", "zip", public_path, "train2")
    shutil.make_archive(public_path / "test2", "zip", public_path, "test2")
    shutil.rmtree(train_output_dir)
    shutil.rmtree(test_output_dir)

    # Create answer file
    answers_df = pd.DataFrame(answer_rows)
    answers_df.to_csv(private_path / "test.csv", index=False)
    assert set(answers_df.columns) == set(["clip", "probability"])

    # Create sample submission file
    sample_submission = answers_df.copy()
    sample_submission["probability"] = 0
    sample_submission.to_csv(public_path / "sampleSubmission.csv", index=False)
    assert set(sample_submission.columns) == set(["clip", "probability"])


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Also creates a secondary validation split in public_val/private_val directories.
    """
    # Data is in train2.zip - we need to unzip it
    shutil.unpack_archive(raw / "train2.zip", raw)

    # Files are named as
    # Train: "YYYYMMDD_HHMMSS_{seconds}_TRAIN{idx}_{label:0,1}.aif"
    # Test: "YYYYMMDD_HHMMSS_{seconds}_Test{idx}.aif"

    # There are 4 days in Train and 3 days in Test
    # In our new dataset, we'll just split Train_old into 2 days for Train and 2 days for Test

    samples_by_date = {}
    n_train_old = 0
    for sample in (raw / "train2").iterdir():
        date = sample.name.split("_")[0]
        if date not in samples_by_date:
            samples_by_date[date] = []
        samples_by_date[date].append(sample)
        n_train_old += 1

    assert len(samples_by_date) == 4, "Expected 4 days in Train_old"
    dates = sorted(list(samples_by_date.keys()))

    # --- 1. Create the Original Split (public/private) ---
    # This split uses the first two days for training and the last two days for testing.
    # The outputs of this step must remain identical to the original script.
    original_train_files = samples_by_date[dates[0]] + samples_by_date[dates[1]]
    original_test_files = samples_by_date[dates[2]] + samples_by_date[dates[3]]

    _create_split(
        train_files=original_train_files,
        test_files=original_test_files,
        public_path=public,
        private_path=private,
    )

    # --- 2. Create the New Validation Split (public_val/private_val) ---
    # This second split takes the original *training* data (the first two days) and
    # splits it again, using the same date-based logic. The first day becomes the
    # new training set, and the second day becomes the new validation (test) set.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    validation_train_files = samples_by_date[dates[0]]
    validation_test_files = samples_by_date[dates[1]]

    _create_split(
        train_files=validation_train_files,
        test_files=validation_test_files,
        public_path=public_val,
        private_path=private_val,
    )

    # Final cleanup of the raw unzipped directory after all processing is complete
    shutil.rmtree(raw / "train2")

    # Final top-level assertions from the original script
    assert (
        len(original_train_files) + len(original_test_files) == n_train_old
    ), f"Expected {n_train_old} total samples in new_train ({len(original_train_files)}) and new_test ({len(original_test_files)})"
    assert (public / "sampleSubmission.csv").exists()
    assert (private / "test.csv").exists()


import logging
import json
from typing import Dict, Any, Optional
from pathlib import Path

from .aide_workflow import AIDEWorkflow
from dsat.services.states.journal import Node, MetricValue
from dsat.common.typing import ExecutionResult
from dsat.benchmark.benchmark import BaseBenchmark

from dsat.services.vdb import VDBService

from dsat.prompts.common import create_draft_prompt 
from dsat.prompts.automind_prompt import create_stepwise_code_prompt, create_stepwise_debug_prompt
from dsat.prompts.aide_prompt import create_improve_prompt, create_debug_prompt

from dsat.utils.context import (
    ContextManager,
    MAX_HISTORY_CHARS,
    MAX_OUTPUT_CHARS,
    summarize_repetitive_logs,
    truncate_output,
)

logger = logging.getLogger(__name__)


class AutoMindWorkflow(AIDEWorkflow):
    """
    Implements the AUTOMIND iterative search algorithm.
    This workflow extends AIDE by incorporating a knowledge base (VDB),
    a self-adaptive coding strategy (one-pass vs. stepwise), and more
    sophisticated context management for complex tasks.
    """
    def __init__(self, operators: Dict[str, Any], services: Dict[str, Any], agent_config: Dict[str, Any], benchmark: Optional[BaseBenchmark] = None):
        """
        Initializes the AutoMindWorkflow, building upon the AIDE base.
        """
        # Initialize base AIDE components, now passing the benchmark instance up.
        super().__init__(operators, services, agent_config, benchmark=benchmark)
        
        self.vdb_service: VDBService = services.get("vdb")
        
        self.complexity_scorer_op = self.operators.get("complexity_scorer")
        self.plan_decomposer_op = self.operators.get("plan_decomposer")
        
        # AutoMind's context manager requires an LLM service to summarize knowledge and history.
        self.context_manager = ContextManager(llm_service=services.get("llm"))

    async def _execute_search_step(self, task_context: Dict, output_path: Path):
        """
        Execute a single step of the AutoMind search loop.
        This overrides the AIDE implementation to add knowledge retrieval, the
        self-adaptive coding strategy, and now uses **grounded validation**.
        
        Args:
            task_context: A dictionary containing the task goal, data report, and I/O instructions.
            output_path: The path where the final output file is expected. Used for grounded validation.
        """
        # 1. Select a node
        parent_node = self._select_node_to_expand()
        
        task_goal = task_context.get('goal_and_data', 'Solve the data science task.')
        io_instructions = task_context.get('io_instructions', 'N/A')

        # 2. Create a prompt
        if parent_node is None:
            # For new drafts, retrieve similar examples from the knowledge base.
            retrieved_knowledge = ""
            if self.vdb_service:
                cases = self.vdb_service.retrieve(task_goal, top_k=2)
                retrieved_knowledge = await self.context_manager.summarize_knowledge(cases, task_goal)
            
            prompt = create_draft_prompt(task_context, self.state.generate_summary(), retrieved_knowledge)
        elif parent_node.is_buggy:
            error_summary = self.context_manager.summarize_error(parent_node.term_out, parent_node.exc_type)
            prompt = create_debug_prompt(task_context, parent_node.code, error_summary, previous_plan=parent_node.plan, memory_summary=self.state.generate_summary())
        else:
            summarized_output = summarize_repetitive_logs(parent_node.term_out)
            prompt = create_improve_prompt(
                task_context, self.state.generate_summary(), parent_node.code,
                parent_node.analysis, previous_plan=parent_node.plan, previous_output=summarized_output
            )

        # 3. Generate initial plan and one-pass code.
        plan, one_pass_code = await self.generate_op(system_prompt=prompt)
        
        # 4. Apply the self-adaptive coding strategy for new drafts.
        use_adaptive = self.complexity_scorer_op and self.plan_decomposer_op
        if use_adaptive and parent_node is None:
            final_code, exec_result = await self._execute_step_adaptively(plan, one_pass_code, task_goal, io_instructions)
        else:
            # For simpler tasks, improvements, or debugging, use the one-pass code.
            final_code = one_pass_code
            exec_result = await self.execute_op(code=final_code, mode="script")

        # 5. Create a new node and absorb the execution result.
        new_node = Node(plan=plan, code=final_code)
        new_node.absorb_exec_result(exec_result)

        if exec_result.success:
            submission_file_in_sandbox = self.sandbox_service.workspace.get_path("sandbox_workdir") / output_path.name
            
            if not submission_file_in_sandbox.exists():
                new_node.is_buggy = True
                new_node.analysis = "Code executed without error, but failed to produce the required output file."
                new_node.metric = MetricValue(value=0.0, maximize=True)
            elif self.benchmark and hasattr(self.benchmark, 'grade'):
                logger.info(f"Performing grounded validation using benchmark grader on '{submission_file_in_sandbox}'...")
                score = await self.benchmark.grade(submission_file_in_sandbox)
                
                if score > 0:
                    new_node.is_buggy = False
                    new_node.metric = MetricValue(value=score, maximize=True)
                    logger.info(f"Grounded validation PASSED. Score: {score:.4f}")
                    review = await self.review_op(prompt_context={
                        "task": task_context, "code": new_node.code, "output": new_node.term_out
                    })
                    new_node.analysis = f"Grounded Score: {score:.4f}. Reviewer Summary: {review.summary}"
                else:
                    new_node.is_buggy = True
                    new_node.metric = MetricValue(value=score, maximize=True)
                    new_node.analysis = "Grounded validation FAILED: The generated submission file was invalid or scored 0.0."
                    logger.warning(f"Grounded validation FAILED. Score: {score}")
            else:
                logger.warning("No benchmark with 'grade' method found. Falling back to unreliable LLM-based review.")
                review = await self.review_op(prompt_context={
                    "task": task_context, "code": new_node.code, "output": new_node.term_out
                })
                new_node.analysis = review.summary
                new_node.is_buggy = review.is_buggy
                new_node.metric = MetricValue(value=review.metric_value, maximize=not review.lower_is_better) if review.metric_value is not None else MetricValue(value=None)
        
        # 8. Add the new node to the search tree state.
        self.state.append(new_node, parent=parent_node)
        logger.info(f"Step {new_node.step} complete. Buggy: {new_node.is_buggy}. Metric: {new_node.metric}.")

    async def _execute_step_adaptively(self, plan: str, one_pass_code: str, task_goal: str, io_instructions: str) -> tuple[str, ExecutionResult]:
        """
        Core of the Self-Adaptive Coding Strategy.
        It scores the complexity of a plan and chooses to either execute the provided
        one-pass code directly or decompose the plan into smaller steps and execute them
        sequentially in a notebook context.
        
        Args:
            plan: The overall plan for the task.
            one_pass_code: The single block of code generated for the entire plan.
            task_goal: The user's primary goal.
            
        Returns:
            A tuple containing the final code (either one-pass or combined steps) and the
            final ExecutionResult.
        """
        final_code = one_pass_code
        
        # 1. Score plan complexity using the dedicated operator.
        score = await self.complexity_scorer_op(plan=plan, task_goal=task_goal)

        # 2. Choose strategy based on the complexity score.
        if score.complexity <= 3:  # Threshold for one-pass vs stepwise
            logger.info("Plan is simple. Executing in one-pass mode.")
            exec_result = await self.execute_op(code=final_code, mode="script")
        else:
            logger.info("Plan is complex. Decomposing and executing in stepwise mode.")
            # 3. Decompose the complex plan into a sequence of smaller tasks.
            decomposed_plan = await self.plan_decomposer_op(plan=plan, task_goal=task_goal)
            
            step_codes = []
            history_steps = []
            final_exec_result = None
            # Get max retries config
            max_step_retries = self.agent_config.get("max_retries", 3)

            async with self.sandbox_service.notebook_executor() as notebook:
                for task in decomposed_plan.tasks:
                    logger.info(f"Executing step {task.task_id}: {task.instruction}")
                    
                    step_succeeded = False
                    current_code = ""
                    step_failure_history = []  # History for the current step

                    # Implement retry loop for robustness
                    for attempt in range(max_step_retries):
                        logger.info(f"Step {task.task_id}, Attempt {attempt + 1}/{max_step_retries}")

                        # Build a concise history of recent steps to provide context for the next step.
                        recent_history_str = self.context_manager.build_history_context(
                            history_steps, 
                            key_order=["task_id", "code", "output"]
                        )
                        
                        if attempt == 0:
                            step_prompt = create_stepwise_code_prompt(task_goal, plan, recent_history_str, task.instruction, io_instructions)
                        else:
                            error_summary = self.context_manager.summarize_error(exec_result.stderr, exec_result.exc_type)
                            step_failure_history.append({
                                "attempt": attempt,
                                "code": truncate_output(current_code, MAX_OUTPUT_CHARS),
                                "error": error_summary
                            })
                            
                            formatted_failure_history = "\n".join([
                                f"--- Attempt {f['attempt']} Failed ---\nCode:\n```python\n{f['code']}\n```\nError: {f['error']}\n---"
                                for f in step_failure_history
                            ])
                            safe_failure_history = truncate_output(formatted_failure_history, MAX_HISTORY_CHARS)
                            
                            step_prompt = create_stepwise_debug_prompt(
                                task_goal, plan, recent_history_str, task.instruction,
                                current_code, safe_failure_history, io_instructions
                            )

                        _, current_code = await self.generate_op(system_prompt=step_prompt)
                        exec_result = await self.execute_op(code=current_code, mode="notebook", executor_context=notebook)
                        
                        if exec_result.success:
                            step_succeeded = True
                            break

                    if not step_succeeded:
                        logger.error(f"Step {task.task_id} failed after {max_step_retries} attempts. Aborting stepwise execution.")
                        final_exec_result = exec_result  # Capture the failed result
                        break

                    step_codes.append(f"# --- Step {task.task_id}: {task.instruction} ---\n{current_code}")
                    # Record the successful step for future context.
                    history_steps.append({
                        "task_id": task.task_id,
                        "code": truncate_output(current_code, MAX_OUTPUT_CHARS),
                        "output": truncate_output(exec_result.stdout, MAX_OUTPUT_CHARS),
                    })
                    final_exec_result = exec_result  # Update with the latest successful result
            
            final_code = "\n\n".join(step_codes)
            exec_result = final_exec_result

        return final_code, exec_result



from pathlib import Path

from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv


def prepare(raw: Path, public: Path, private: Path):

    # Create train and test splits from train set
    old_train = read_csv(raw / "train.csv")
    new_train, new_test = train_test_split(old_train, test_size=0.1, random_state=0)

    # Create sample submission
    sample_submission = new_test.copy()
    sample_submission["EC1"] = 0.5
    sample_submission["EC2"] = 0.5
    sample_submission.drop(
        sample_submission.columns.difference(["id", "EC1", "EC2"]), axis=1, inplace=True
    )
    sample_submission.to_csv(public / "sample_submission.csv", index=False)

    # Create private files
    new_test.to_csv(private / "test.csv", index=False)

    # Create public files visible to agents
    new_train.to_csv(public / "train.csv", index=False)
    new_test.drop(["EC1", "EC2", "EC3", "EC4", "EC5", "EC6"], axis=1, inplace=True)
    new_test.to_csv(public / "test.csv", index=False)

    # Checks
    assert new_test.shape[1] == 32, "Public test set should have 32 columns"
    assert new_train.shape[1] == 38, "Public train set should have 38 columns"
    assert len(new_train) + len(new_test) == len(
        old_train
    ), "Length of new_train and new_test should equal length of old_train"



"""
Package Detector - Detects available Python packages in the current environment.

This module provides utilities to scan the current Python environment and
detect installed packages, making this information available to the agent
for better code generation.
"""

import importlib.metadata as metadata
import logging
from typing import Dict, List, Optional
from pathlib import Path
import yaml
import json

logger = logging.getLogger(__name__)


class PackageDetector:
    """
    Detects and manages information about available Python packages.

    This class scans the current Python environment to detect installed
    packages and can save/load this information from config files.
    """

    # Common data science packages to highlight
    DATA_SCIENCE_PACKAGES = {
        'pandas', 'numpy', 'scipy', 'scikit-learn', 'sklearn',
        'matplotlib', 'seaborn', 'plotly', 'bokeh',
        'xgboost', 'lightgbm', 'catboost', 'torch', 'tensorflow',
        'keras', 'fastai', 'transformers', 'datasets',
        'statsmodels', 'nltk', 'spacy', 'gensim',
        'cv2', 'PIL', 'pillow', 'imageio',
        'requests', 'beautifulsoup4', 'bs4', 'scrapy',
        'sqlalchemy', 'pymongo', 'psycopg2',
        'polars', 'pyarrow', 'dask'
    }

    def __init__(self, include_stdlib: bool = False):
        """
        Initialize PackageDetector.

        Args:
            include_stdlib: Whether to include standard library packages
        """
        self.include_stdlib = include_stdlib
        self._packages_cache: Optional[Dict[str, str]] = None

    def detect_packages(self, force_refresh: bool = False) -> Dict[str, str]:
        """
        Detect all installed packages in the current environment.

        Args:
            force_refresh: Force re-detection even if cached

        Returns:
            Dictionary mapping package names to their versions
        """
        if self._packages_cache is not None and not force_refresh:
            return self._packages_cache

        packages = {}

        try:
            # Use importlib.metadata to get installed packages
            for dist in metadata.distributions():
                name = dist.metadata['Name'].lower()
                version = dist.version

                # Filter out standard library if not requested
                if not self.include_stdlib and self._is_stdlib(name):
                    continue

                packages[name] = version

            self._packages_cache = packages
            logger.info(f"Detected {len(packages)} packages in environment")

        except Exception as e:
            logger.error(f"Error detecting packages: {e}")
            # Fallback to basic package list
            packages = self._get_fallback_packages()

        return packages

    def _is_stdlib(self, package_name: str) -> bool:
        """
        Check if a package is part of the Python standard library.

        Args:
            package_name: Name of the package

        Returns:
            True if it's a standard library package
        """
        # List of common stdlib packages
        stdlib_packages = {
            'asyncio', 'concurrent', 'csv', 'datetime', 'decimal',
            'email', 'json', 'logging', 'math', 'os', 'pathlib',
            'random', 're', 'socket', 'sqlite3', 'sys', 'time',
            'urllib', 'uuid', 'warnings', 'collections', 'itertools',
            'functools', 'typing', 'dataclasses', 'enum', 'typing_extensions'
        }

        return package_name in stdlib_packages

    def _get_fallback_packages(self) -> Dict[str, str]:
        """
        Fallback method to get basic packages if metadata detection fails.

        Returns:
            Dictionary of package names to versions
        """
        packages = {}

        # Try to import common packages
        common_packages = [
            'pandas', 'numpy', 'scipy', 'sklearn', 'matplotlib',
            'seaborn', 'xgboost', 'lightgbm', 'torch', 'tensorflow'
        ]

        for pkg in common_packages:
            try:
                mod = __import__(pkg)
                version = getattr(mod, '__version__', 'unknown')
                packages[pkg] = version
            except ImportError:
                pass

        return packages

    def get_data_science_packages(self) -> Dict[str, str]:
        """
        Get only data science related packages.

        Returns:
            Dictionary of data science package names to versions
        """
        all_packages = self.detect_packages()
        ds_packages = {
            name: version
            for name, version in all_packages.items()
            if name in self.DATA_SCIENCE_PACKAGES
        }
        return ds_packages

    def format_as_context(self, packages: Optional[Dict[str, str]] = None) -> str:
        """
        Format package information as a context string for the agent.

        Args:
            packages: Package dictionary (uses detected if None)

        Returns:
            Formatted string describing available packages
        """
        if packages is None:
            packages = self.detect_packages()

        if not packages:
            return "No package information available."

        lines = [
            "Available Python packages in the current environment:",
            ""
        ]

        # Group by category
        ds_packages = self.get_data_science_packages()

        if ds_packages:
            lines.append("Data Science & ML Packages:")
            for name, version in sorted(ds_packages.items()):
                lines.append(f"  - {name} ({version})")
            lines.append("")

        # Other packages (limit to most common)
        other_packages = {
            name: version
            for name, version in packages.items()
            if name not in ds_packages
        }

        # Show top 20 other packages alphabetically
        if other_packages:
            lines.append("Other Available Packages (showing top 20):")
            for name, version in sorted(other_packages.items())[:20]:
                lines.append(f"  - {name} ({version})")

            if len(other_packages) > 20:
                lines.append(f"  ... and {len(other_packages) - 20} more packages")

        return "\n".join(lines)

    def save_to_config(
        self,
        config_path: Path,
        packages: Optional[Dict[str, str]] = None,
        data_science_only: bool = True
    ):
        """
        Save detected packages to a config file.

        Args:
            config_path: Path to config.yaml file
            packages: Package dictionary to save (detects if None)
            data_science_only: If True, only save data science packages (default: True)
        """
        if packages is None:
            packages = self.detect_packages()

        # Filter to only data science packages if requested
        if data_science_only:
            packages = {
                name: version
                for name, version in packages.items()
                if name in self.DATA_SCIENCE_PACKAGES
            }
            logger.info(f"Filtered to {len(packages)} data science packages")

        try:
            # Read existing config
            if config_path.exists():
                with open(config_path, 'r') as f:
                    config = yaml.safe_load(f) or {}
            else:
                config = {}

            # Add packages to config
            config['available_packages'] = {
                'enabled': True,
                'packages': packages,
                'last_updated': str(Path(__file__).stat().st_mtime)
            }

            # Write back
            with open(config_path, 'w') as f:
                yaml.dump(config, f, default_flow_style=False, sort_keys=False)

            logger.info(f"Saved {len(packages)} packages to {config_path}")

        except Exception as e:
            logger.error(f"Error saving packages to config: {e}")

    def load_from_config(self, config_path: Path) -> Optional[Dict[str, str]]:
        """
        Load packages from a config file.

        Args:
            config_path: Path to config.yaml file

        Returns:
            Package dictionary if found, None otherwise
        """
        try:
            if not config_path.exists():
                return None

            with open(config_path, 'r') as f:
                config = yaml.safe_load(f) or {}

            pkg_config = config.get('available_packages', {})

            if not pkg_config.get('enabled', True):
                logger.info("Package detection is disabled in config")
                return None

            packages = pkg_config.get('packages', {})

            if packages:
                logger.info(f"Loaded {len(packages)} packages from {config_path}")
                return packages

            return None

        except Exception as e:
            logger.error(f"Error loading packages from config: {e}")
            return None


def detect_and_save_packages(config_path: Optional[Path] = None) -> Dict[str, str]:
    """
    Convenience function to detect packages and save to config.

    Args:
        config_path: Path to config.yaml (uses default if None)

    Returns:
        Dictionary of detected packages
    """
    detector = PackageDetector()

    if config_path is None:
        # Try to find config.yaml in current directory or parent
        current_path = Path.cwd()
        while current_path != current_path.parent:
            potential_config = current_path / 'config.yaml'
            if potential_config.exists():
                config_path = potential_config
                break
            current_path = current_path.parent

        if config_path is None:
            config_path = Path.cwd() / 'config.yaml'

    packages = detector.detect_packages()
    detector.save_to_config(config_path, packages)

    return packages


if __name__ == "__main__":
    # Test the detector
    logging.basicConfig(level=logging.INFO)

    detector = PackageDetector()
    packages = detector.detect_packages()

    print("\n" + "="*70)
    print("Detected Packages")
    print("="*70)
    print(detector.format_as_context(packages))

    print("\n" + "="*70)
    print("Data Science Packages")
    print("="*70)
    ds_packages = detector.get_data_science_packages()
    for name, version in sorted(ds_packages.items()):
        print(f"  {name}: {version}")

    # Save to config
    config_path = Path("config.yaml")
    detector.save_to_config(config_path, packages)
    print(f"\n✓ Saved to {config_path}")



import shutil
from pathlib import Path

from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv

from .classes import CLASSES


def prepare(raw: Path, public: Path, private: Path):
    # Create train, test from train split
    old_train = read_csv(raw / "train.csv")
    new_train, new_test = train_test_split(old_train, test_size=0.1, random_state=0)

    old_train_annotations = read_csv(raw / "train_annotations.csv")
    old_train_uids = old_train_annotations["StudyInstanceUID"]
    new_train_uids = new_train["StudyInstanceUID"]
    is_in_new_train = old_train_uids.isin(new_train_uids)

    new_train_annotations = old_train_annotations[is_in_new_train]

    (public / "train").mkdir(exist_ok=True)
    (public / "test").mkdir(exist_ok=True)

    for file_id in new_train["StudyInstanceUID"]:
        shutil.copyfile(
            src=raw / "train" / f"{file_id}.jpg",
            dst=public / "train" / f"{file_id}.jpg",
        )

    for file_id in new_test["StudyInstanceUID"]:
        shutil.copyfile(
            src=raw / "train" / f"{file_id}.jpg",
            dst=public / "test" / f"{file_id}.jpg",
        )

    assert len(list(public.glob("train/*.jpg"))) == len(
        new_train
    ), f"Expected {len(new_train)} files in public train, got {len(list(public.glob('train/*.jpg')))}"
    assert len(list(public.glob("test/*.jpg"))) == len(
        new_test
    ), f"Expected {len(new_test)} files in public test, got {len(list(public.glob('test/*.jpg')))}"

    # Create a sample submission file
    submission_df = new_test[["StudyInstanceUID"] + CLASSES]
    submission_df[CLASSES] = 0

    # Copy over files
    new_train.to_csv(public / "train.csv", index=False)
    new_train_annotations.to_csv(public / "train_annotations.csv", index=False)
    new_test.to_csv(private / "test.csv", index=False)
    submission_df.to_csv(public / "sample_submission.csv", index=False)



import shutil
from pathlib import Path

# New import required for the data splitting
from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv


def prepare(raw: Path, public: Path, private: Path):
    """
    Prepares the data for the competition.

    This script performs two main operations:
    1.  Processes the original raw data into `public` and `private` directories
        for the official test set, ensuring the output is identical to the
        original script's output.
    2.  Creates a new validation split from the original training data. It splits
        the original `train.csv` into a new, smaller training set (`train_val`)
        and a validation set (`test_val`). The results are saved in parallel
        `public_val` and `private_val` directories with a mirrored file structure.
    """
    # --- Part 1: Original Data Preparation (UNCHANGED OUTPUT) ---
    # This block ensures the 'public' and 'private' directories are populated
    # exactly as they were in the original script.

    # Ensure original output directories exist
    public.mkdir(parents=True, exist_ok=True)
    private.mkdir(parents=True, exist_ok=True)

    # Create test and test without labels
    test = read_csv(raw / "test_private_expanded.csv")
    test_without_labels = test[["id", "comment_text"]]

    # Copy over files
    shutil.copy(raw / "test_private_expanded.csv", private / "test.csv")
    shutil.copy(raw / "train.csv", public / "train.csv")
    test_without_labels.to_csv(public / "test.csv", index=False)

    # Create a sample submission file
    submission_df = test_without_labels.copy()
    submission_df = submission_df.drop(columns=["comment_text"])
    submission_df["prediction"] = 0.0
    submission_df.to_csv(public / "sample_submission.csv", index=False)

    # Checks
    assert test.shape == (
        len(test),
        45,
    ), "test.csv should have 45 columns as per raw data"
    assert test_without_labels.shape == (
        len(test_without_labels),
        2,
    ), "test.csv should have 2 columns: id, comment_text"

    assert submission_df.shape == (
        len(test),
        2,
    ), "sample_submission.csv should have 2 columns: id, prediction"

    # --- Part 2: New Validation Set Creation ---
    # This block creates a new split from the original training data to form
    # a new, smaller training set and a validation set. Outputs are saved
    # to 'public_val' and 'private_val' directories.

    # Define and create the new parallel directories for the validation split
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"
    public_val.mkdir(parents=True, exist_ok=True)
    private_val.mkdir(parents=True, exist_ok=True)

    # Load the full original training data, which will be split
    full_train_df = read_csv(raw / "train.csv")

    # The size of the new validation set ('test_val') should be the same
    # as the size of the original test set to replicate the split ratio.
    test_set_size = len(test)

    # Split the full training data into a new training and validation set.
    # A fixed random_state is used to ensure the split is deterministic.
    train_val, test_val = train_test_split(
        full_train_df, test_size=test_set_size, random_state=42
    )

    # Create the public version of the validation set (without labels)
    test_val_without_labels = test_val[["id", "comment_text"]]

    # Save the new split data, mirroring the original file structure and names.
    # private_val contains the ground truth for the new validation set
    test_val.to_csv(private_val / "test.csv", index=False)

    # public_val contains the new smaller training set and the unlabeled validation set
    train_val.to_csv(public_val / "train.csv", index=False)
    test_val_without_labels.to_csv(public_val / "test.csv", index=False)

    # Create a sample submission file for the new validation set
    submission_df_val = test_val_without_labels.copy()
    submission_df_val = submission_df_val.drop(columns=["comment_text"])
    submission_df_val["prediction"] = 0.0
    submission_df_val.to_csv(public_val / "sample_submission.csv", index=False)

    # Add checks for the new validation split dataframes
    assert test_val.shape == (
        len(test_val),
        full_train_df.shape[1],
    ), f"private_val/test.csv should have {full_train_df.shape[1]} columns"
    assert test_val_without_labels.shape == (
        len(test_val_without_labels),
        2,
    ), "public_val/test.csv should have 2 columns: id, comment_text"
    assert submission_df_val.shape == (
        len(test_val),
        2,
    ), "public_val/sample_submission.csv should have 2 columns: id, prediction"
    assert len(train_val) + len(test_val) == len(
        full_train_df
    ), "Split should conserve all data from original train set"


import shutil
from pathlib import Path

from sklearn.model_selection import train_test_split
from tqdm.auto import tqdm

from mlebench.utils import compress, extract, read_csv


def prepare(raw: Path, public: Path, private: Path):

    old_train = read_csv(raw / "train.csv")

    # 4000 / (4000 + 17500) -> test_size is ~0.19
    new_train, new_test = train_test_split(old_train, test_size=0.19, random_state=0)

    new_train.to_csv(public / "train.csv", index=False)
    new_test.to_csv(private / "test.csv", index=False)

    sample_submission = new_test.copy()
    sample_submission["has_cactus"] = 0.5
    sample_submission.to_csv(public / "sample_submission.csv", index=False)

    # need to split the train.zip into train.zip and test.zip; to do so need to extract first
    extract(raw / "train.zip", raw)

    # copy
    (public / "train").mkdir(parents=True, exist_ok=True)
    for image_id in tqdm(new_train["id"], desc="Copying train images", total=len(new_train)):
        shutil.copy(raw / "train" / image_id, public / "train" / image_id)

    (public / "test").mkdir(parents=True, exist_ok=True)
    for image_id in tqdm(new_test["id"], desc="Copying test images", total=len(new_test)):
        shutil.copy(raw / "train" / image_id, public / "test" / image_id)

    # and then recompress
    compress(public / "train", public / "train.zip")
    compress(public / "test", public / "test.zip")

    # and cleanup
    shutil.rmtree(public / "train")
    shutil.rmtree(public / "test")
    shutil.rmtree(raw / "train")

    # checks
    assert (public / "train.zip").exists(), "public/train.zip should exist"
    assert (public / "test.zip").exists(), "public/test.zip should exist"
    assert not (raw / "train").exists(), "raw/train/ should not exist"

    assert len(new_train) + len(new_test) == len(
        old_train
    ), "The lengths of the splits should add up to the original"
    assert len(new_train) > len(new_test), "The train set should be larger than the test set"
    assert len(new_test) == len(
        sample_submission
    ), "The test set should match the sample submission"

    assert (
        new_train.columns.tolist()
        == new_test.columns.tolist()
        == old_train.columns.tolist()
        == sample_submission.columns.tolist()
    ), "All dataframes should have the same columns, i.e. ['id', 'has_cactus']"

    assert set(new_train["id"]).isdisjoint(new_test["id"]), "Train and test ids should not overlap"
    assert set(new_test["id"]) == set(
        sample_submission["id"]
    ), "Test and sample_submission ids should match"

    assert new_train["id"].nunique() == len(new_train), "There should be no duplicate ids in train"
    assert new_test["id"].nunique() == len(new_test), "There should be no duplicate ids in test"



import shutil
from pathlib import Path

from mlebench.utils import read_csv


def prepare(raw: Path, public: Path, private: Path):
    # the test set labels are provided so we can just copy things without doing the split ourselves
    shutil.copy(raw / "train.csv", public / "train.csv")
    shutil.copy(raw / "test_with_solutions.csv", private / "test.csv")

    # the public test currently online is for the verification stage, which we are ignoring
    # (we are ignoring because there is some train/test leakage and unclear how this factors in leaderboard)
    # we can recover the original public test set by removing the labels from the private test set
    # can make the gold submission while we're at it
    gold_submission = read_csv(private / "test.csv")
    gold_submission = gold_submission[["Insult", "Date", "Comment"]]
    gold_submission.to_csv(private / "gold_submission.csv", index=False)

    public_test = gold_submission.drop(columns=["Insult"]).copy()
    public_test.to_csv(public / "test.csv", index=False)

    # finally, we also make our own sample_submission, same reasons as public test
    # but match the format of what's online
    sample_submission = gold_submission.copy()
    sample_submission["Insult"] = 0
    sample_submission.to_csv(public / "sample_submission_null.csv", index=False)

    # checks
    public_train = read_csv(public / "train.csv")
    public_test = read_csv(public / "test.csv")
    private_test = read_csv(private / "test.csv")

    # no `Id` column in train, so we check comment content instead
    assert public_train.columns.to_list() == [
        "Insult",
        "Date",
        "Comment",
    ], "Train columns should be Insult, Date, Comment"
    assert public_test.columns.to_list() == [
        "Date",
        "Comment",
    ], "Test columns should be Date, Comment"
    assert sample_submission.columns.to_list() == [
        "Insult",
        "Date",
        "Comment",
    ], "Sample submission columns should be Insult, Date, Comment"
    assert gold_submission.columns.to_list() == [
        "Insult",
        "Date",
        "Comment",
    ], "Gold submission columns should be Insult, Date, Comment"
    assert private_test.columns.to_list() == [
        "Insult",
        "Date",
        "Comment",
        "Usage",
    ], "Private test columns should be Insult, Date, Comment, Usage"

    assert set(public_train["Comment"]).isdisjoint(
        set(public_test["Comment"])
    ), "None of the test comments should be in the train comments"
    assert public_test.equals(
        private_test.drop(columns=["Insult", "Usage"])
    ), "Public test should be identical to private test, modulo the Insult and Usage columns"
    assert set(public_test["Comment"]) == set(
        sample_submission["Comment"]
    ), "Public test and sample submission should have the same Comments"
    assert set(public_test["Comment"]) == set(
        gold_submission["Comment"]
    ), "Public test and gold submission should have the same Comments"



import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

from mlebench.utils import extract, read_csv


def _create_split(
    df_in: pd.DataFrame,
    public_path: Path,
    private_path: Path,
    test_size: float,
    random_state: int,
) -> pd.DataFrame:
    """
    Helper function to perform a data split and write files to specified directories.

    Args:
        df_in: The input DataFrame to split.
        public_path: The directory for public-facing files (train set, test features).
        private_path: The directory for private files (test answers).
        test_size: The proportion of the dataset to allocate to the test set.
        random_state: The seed for the random number generator.

    Returns:
        The DataFrame for the newly created training set.
    """
    # Create output directories if they don't exist
    public_path.mkdir(exist_ok=True, parents=True)
    private_path.mkdir(exist_ok=True, parents=True)

    # Create train, test from the input dataframe
    new_train, answers = train_test_split(
        df_in, test_size=test_size, random_state=random_state
    )
    new_test = answers[["id", "prompt", "response_a", "response_b"]].copy()

    sample_submission = answers[["id"]].copy()
    sample_submission["winner_model_a"] = 0.3333333333333333
    sample_submission["winner_model_b"] = 0.3333333333333333
    sample_submission["winner_tie"] = 0.3333333333333333

    # Checks
    assert len(new_train) + len(new_test) == len(
        df_in
    ), f"New train and test should have the same number of rows as the original dataframe"
    assert set(new_train["id"]).isdisjoint(
        set(new_test["id"])
    ), f"New train and test should have no overlapping ids"
    assert new_test.columns.tolist() == [
        "id",
        "prompt",
        "response_a",
        "response_b",
    ], f"New test should have columns id, prompt, response_a, response_b"
    assert sample_submission.columns.tolist() == [
        "id",
        "winner_model_a",
        "winner_model_b",
        "winner_tie",
    ], f"Sample submission should have columns id, winner_model_a, winner_model_b, winner_tie"
    assert (
        new_train.columns.tolist() == df_in.columns.tolist()
    ), f"New train should have the same columns as the original dataframe"

    # Write CSVs
    answers.to_csv(private_path / "answers.csv", index=False)
    new_train.to_csv(public_path / "train.csv", index=False)
    new_test.to_csv(public_path / "test.csv", index=False)
    sample_submission.to_csv(public_path / "sample_submission.csv", index=False)

    return new_train


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Also creates a secondary validation split (public_val, private_val) for local testing.
    """

    # --- Stage 1: Create the original competition split (train/test) ---
    # This block generates the primary `public` and `private` directories.
    # Its outputs MUST remain identical to the original script's outputs.
    old_train_df = read_csv(raw / "train.csv")
    train_for_val_split = _create_split(
        df_in=old_train_df,
        public_path=public,
        private_path=private,
        test_size=0.1,
        random_state=0,
    )

    # --- Stage 2: Create the new validation split (train_val/test_val) ---
    # This block takes the training set from Stage 1 and splits it again
    # to create a new, smaller training set and a validation set.
    # The outputs are saved to parallel `public_val` and `private_val` directories.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Calculate the test_size needed to make the new validation set (`test_val`)
    # have the same number of samples as the original test set from Stage 1.
    # Original test size = 0.1 * total. New train size = 0.9 * total.
    # We need a fraction `p` such that p * (0.9 * total) = 0.1 * total.
    # p = 0.1 / 0.9 = 1/9.
    val_test_size = 1 / 9

    _create_split(
        df_in=train_for_val_split,
        public_path=public_val,
        private_path=private_val,
        test_size=val_test_size,
        random_state=0,  # Use same random state for consistency
    )


import pandas as pd

from mlebench.grade_helpers import InvalidSubmissionError
from mlebench.metrics import mean_average_precision_at_k


def prepare_for_metric(submission: pd.DataFrame, answers: pd.DataFrame) -> dict:
    # Answers checks
    assert set(answers.columns) == {
        "customer_id",
        "prediction",
    }, "Answers DataFrame must have 'customer_id' and 'prediction' columns."

    # Submission checks
    if set(submission.columns) != {"customer_id", "prediction"}:
        raise InvalidSubmissionError(
            "Submission DataFrame must have 'customer_id' and 'prediction' columns."
        )
    if not (set(submission["customer_id"]) >= set(answers["customer_id"])):
        raise InvalidSubmissionError(
            "Submission customer_id must be a superset of answers customer_id"
        )

    # Filter the submission to only consider the customer_ids that exist in answers
    submission = submission[submission["customer_id"].isin(answers["customer_id"])]

    # Sort to ensure we're grading the right rows
    submission = submission.sort_values("customer_id").reset_index(drop=True)
    answers = answers.sort_values("customer_id").reset_index(drop=True)

    y_true = answers["prediction"].astype(str).str.split(" ").apply(set).tolist()
    y_pred = submission["prediction"].astype(str).str.split(" ").tolist()

    return {"actual": y_true, "predicted": y_pred}


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    # Prepare the data for metric calculation
    prepped = prepare_for_metric(submission, answers)
    return mean_average_precision_at_k(
        actual=prepped["actual"], predicted=prepped["predicted"], k=12
    )



import shutil
from pathlib import Path

import pandas as pd
from tqdm.auto import tqdm

from mlebench.competitions.utils import get_ids_from_tf_records
from mlebench.utils import get_logger

logger = get_logger(__name__)


def prepare(raw: Path, public: Path, private: Path):
    # split on the TFRecord files.
    # There's 33126 train samples; 16 TFRecord files with 2071 samples each
    # so we take 2 arbitrary TF record files as our test set (4142 samples is ~ 10% of data)
    DEV = False
    if DEV:
        cutoff_index = 10000
    else:
        cutoff_index = None

    old_train = pd.read_csv(raw / "train.csv")[:cutoff_index]

    test_tf_records = {"train00-2071.tfrec", "train06-2071.tfrec"}
    # parse the IDs from the test tf records
    test_ids = []
    for tfrec in tqdm(test_tf_records, desc="Splitting test ids from train TFRecords"):
        test_ids.extend(get_ids_from_tf_records(raw / "tfrecords" / tfrec))

    old_train["split"] = "train"
    old_train.loc[old_train["image_name"].isin(test_ids), "split"] = "test"

    new_train = old_train[old_train["split"] == "train"].drop(columns=["split"])
    new_test = old_train[old_train["split"] == "test"].drop(columns=["split"])
    new_test_without_labels = new_test.copy()[
        ["image_name", "patient_id", "sex", "age_approx", "anatom_site_general_challenge"]
    ]

    # match format of sample submission
    new_test = new_test[["image_name", "target"]]

    # sample submission
    sample_submission = new_test.copy()
    sample_submission["target"] = 0

    # save the CSVs
    new_train.to_csv(public / "train.csv", index=False)
    new_test_without_labels.to_csv(public / "test.csv", index=False)
    sample_submission.to_csv(public / "sample_submission.csv", index=False)
    new_test.to_csv(private / "test.csv", index=False)

    # split raw train files to into appropriate prepared/public/test/train directories
    # the files themselves do not contain target metadata so we are free to move them around
    # DICOMs and JPEGs
    (public / "train").mkdir(parents=True, exist_ok=True)
    (public / "jpeg" / "train").mkdir(parents=True, exist_ok=True)
    for image_name in tqdm(new_train["image_name"], desc="Train Images", total=len(new_train)):
        dcm_file = raw / "train" / f"{image_name}.dcm"
        jpg_file = raw / "jpeg" / "train" / f"{image_name}.jpg"
        shutil.copy(dcm_file, public / "train" / f"{image_name}.dcm")
        shutil.copy(jpg_file, public / "jpeg" / "train" / f"{image_name}.jpg")
    (public / "test").mkdir(parents=True, exist_ok=True)
    (public / "jpeg" / "test").mkdir(parents=True, exist_ok=True)
    for image_name in tqdm(new_test["image_name"], desc="Test Images", total=len(new_test)):
        dcm_file = raw / "train" / f"{image_name}.dcm"
        jpg_file = raw / "jpeg" / "train" / f"{image_name}.jpg"
        shutil.copy(dcm_file, public / "test" / f"{image_name}.dcm")
        shutil.copy(jpg_file, public / "jpeg" / "test" / f"{image_name}.jpg")

    # TFRecords
    train_count = 0
    test_count = 0
    tfrecords_dest_path = public / "tfrecords"
    tfrecords_dest_path.mkdir(parents=True, exist_ok=True)
    for file in tqdm(
        sorted((raw / "tfrecords").glob("train*.tfrec")), desc="Copying TFRecord files"
    ):
        record_count = file.stem.split("-")[1]  # i.e. get 2071 from train00-2071
        if file.name in test_tf_records:
            shutil.copy(file, tfrecords_dest_path / f"test{test_count:02d}-{record_count}.tfrec")
            test_count += 1
        else:
            shutil.copy(file, tfrecords_dest_path / f"train{train_count:02d}-{record_count}.tfrec")
            train_count += 1

    logger.info("Running asserts...")
    assert len(list(public.glob("train/*.dcm"))) == len(new_train), "Train DICOM count mismatch"
    assert len(list(public.glob("test/*.dcm"))) == len(new_test), "Test DICOM count mismatch"
    assert len(list(public.glob("jpeg/train/*.jpg"))) == len(new_train), "Train JPEG count mismatch"
    assert len(list(public.glob("jpeg/test/*.jpg"))) == len(new_test), "Test JPEG count mismatch"

    assert not set(new_train["image_name"]).intersection(
        new_test["image_name"]
    ), "Train/Test overlap"

    assert len(sample_submission) == len(new_test), "Sample submission length mismatch"
    assert (
        sample_submission["image_name"]
        .sort_values()
        .reset_index(drop=True)
        .equals(new_test["image_name"].sort_values().reset_index(drop=True))
    ), "Sample submission IDs mismatch"

    # check columns
    assert new_test_without_labels.columns.tolist() == [
        "image_name",
        "patient_id",
        "sex",
        "age_approx",
        "anatom_site_general_challenge",
    ], "new_test_without_labels columns mismatch"
    assert new_train.columns.tolist() == [
        "image_name",
        "patient_id",
        "sex",
        "age_approx",
        "anatom_site_general_challenge",
        "diagnosis",
        "benign_malignant",
        "target",
    ], "new_train columns mismatch"
    assert new_test.columns.tolist() == ["image_name", "target"], "new_test columns mismatch"
    assert sample_submission.columns.tolist() == [
        "image_name",
        "target",
    ], "sample_submission columns mismatch"



import shutil
from pathlib import Path

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import read_csv


def _perform_split(
    df: pd.DataFrame, target_test_size: float, random_seed: int
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Performs the custom splitting logic on a given dataframe.

    This logic ensures that for IDs with 2 or more images, one is guaranteed
    to be in the train set and one in the test set. The remaining data is
    split randomly.

    Args:
        df: The input dataframe with "Image" and "Id" columns.
        target_test_size: The approximate fraction of data to allocate to the test set.
        random_seed: The random seed for reproducibility.

    Returns:
        A tuple containing the train dataframe and the test dataframe.
    """
    # Make a copy to avoid modifying the original dataframe
    data_to_split = df.copy()
    data_to_split["split"] = "undecided"

    # seeded random generator for numpy
    np_rng = np.random.default_rng(random_seed)

    # ensure each id occurs in train and test set at least once
    # when there's only one image for an id, goes randomly to train or test
    whale_ids = data_to_split["Id"].unique()
    for whale_id in whale_ids:
        whale_images = data_to_split[data_to_split["Id"] == whale_id]
        if len(whale_images) >= 2:
            # randomly assign one of these to train and one to test
            selected = whale_images.sample(2, random_state=random_seed)
            data_to_split.loc[selected.index[0], "split"] = "train"
            data_to_split.loc[selected.index[1], "split"] = "test"
        else:
            # randomly assign this one image to train or test
            data_to_split.loc[whale_images.index[0], "split"] = np_rng.choice(
                ["train", "test"],
                replace=False,
                p=[1 - target_test_size, target_test_size],
            )

    # split the remaining data
    remaining_data = data_to_split[data_to_split["split"] == "undecided"]
    if not remaining_data.empty:
        train, test = train_test_split(
            remaining_data, test_size=target_test_size, random_state=random_seed
        )
        data_to_split.loc[train.index, "split"] = "train"
        data_to_split.loc[test.index, "split"] = "test"

    # finally, can split out into separate dataframes
    train_df = data_to_split[data_to_split["split"] == "train"].drop(
        columns=["split"]
    )
    test_df = data_to_split[data_to_split["split"] == "test"].drop(columns=["split"])

    return train_df.copy(), test_df.copy()


def _write_output_files(
    train_df: pd.DataFrame,
    test_df: pd.DataFrame,
    public_dir: Path,
    private_dir: Path,
    raw_dir: Path,
):
    """
    Writes all required output files for a given train/test split.
    This includes CSVs, images, and a sample submission file.
    """
    public_dir.mkdir(exist_ok=True)
    private_dir.mkdir(exist_ok=True)

    # Make a copy of the test dataframe to create the final answers
    answers = test_df.copy()

    # If a whale Id is only in the test set, it should be labeled as new_whale instead
    ids_in_test_but_not_train = set(answers["Id"]) - set(train_df["Id"])
    answers.loc[answers["Id"].isin(ids_in_test_but_not_train), "Id"] = "new_whale"

    # Create sample submission
    sample_submission = answers.copy()
    sample_submission["Id"] = "new_whale w_1287fbc w_98baff9 w_7554f44 w_1eafe46"

    # Checks
    assert len(answers) == len(
        sample_submission
    ), "Answers and sample submission should have the same length"
    assert train_df.shape[1] == 2, "Train should have exactly 2 columns"
    assert sample_submission.shape[1] == 2, "Sample submission should have exactly 2 columns"
    assert answers.shape[1] == 2, "Answers should have exactly 2 columns"
    assert (
        "new_whale" in answers["Id"].values
    ), "Answers should contain at least some values with 'new_whale' in the 'Id' column"
    assert len(train_df) + len(answers) == len(train_df) + len(
        test_df
    ), "The combined length of train_df and answers should equal their original combined length"

    # Write CSVs
    answers.to_csv(private_dir / "test.csv", index=False)
    train_df.to_csv(public_dir / "train.csv", index=False)
    sample_submission.to_csv(public_dir / "sample_submission.csv", index=False)

    # Copy over files
    (public_dir / "test").mkdir(exist_ok=True)
    (public_dir / "train").mkdir(exist_ok=True)

    for file_id in tqdm(train_df["Image"], desc=f"Copying train images to {public_dir}"):
        shutil.copyfile(
            src=raw_dir / "train" / f"{file_id}",
            dst=public_dir / "train" / f"{file_id}",
        )

    for file_id in tqdm(answers["Image"], desc=f"Copying test images to {public_dir}"):
        shutil.copyfile(
            src=raw_dir / "train" / f"{file_id}",
            dst=public_dir / "test" / f"{file_id}",
        )

    # File checks
    train_files = list(public_dir.glob("train/*.jpg"))
    test_files = list(public_dir.glob("test/*.jpg"))
    assert len(train_files) == len(
        train_df
    ), f"Train dir {public_dir / 'train'} should have the same number of images as the length of its train set"
    assert len(test_files) == len(
        answers
    ), f"Test dir {public_dir / 'test'} should have the same number of images as the length of its test set"
    assert not set(train_files) & set(test_files), "Train and test files should be distinct"


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Additionally, it creates a second parallel split (train_val, test_val) for validation purposes.
    """
    old_train = read_csv(raw / "train.csv")
    target_test_size = 0.1
    random_seed = 0

    # --- Step 1: Create the original train/test split ---
    # This split generates the main competition data.
    new_train, answers = _perform_split(
        df=old_train, target_test_size=target_test_size, random_seed=random_seed
    )

    # Write the original output files. This part is unchanged in its output.
    _write_output_files(
        train_df=new_train,
        test_df=answers,
        public_dir=public,
        private_dir=private,
        raw_dir=raw,
    )

    # --- Step 2: Create the new validation split ---
    # This takes the `new_train` set from the first split and splits it again
    # to create a validation set (`test_val`) of a similar size to the original `answers`.
    # The new, smaller training set is `train_val`.

    # Define paths for the new validation directories
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # To get a validation test set of size ~0.1*N from a training set of size ~0.9*N,
    # the new test_size must be 0.1/0.9.
    val_target_test_size = target_test_size / (1 - target_test_size)

    # Perform the second split on the `new_train` data using the same logic and seed.
    train_val, test_val = _perform_split(
        df=new_train,
        target_test_size=val_target_test_size,
        random_seed=random_seed,
    )

    # Write the validation output files to the new directories.
    _write_output_files(
        train_df=train_val,
        test_df=test_val,
        public_dir=public_val,
        private_dir=private_val,
        raw_dir=raw,
    )


import pandas as pd
from sklearn.metrics import roc_auc_score

from mlebench.competitions.utils import prepare_for_auroc_metric


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    roc_auc_inputs = prepare_for_auroc_metric(
        submission=submission, answers=answers, id_col="id", target_col="has_cactus"
    )
    return roc_auc_score(y_true=roc_auc_inputs["y_true"], y_score=roc_auc_inputs["y_score"])



import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm.auto import tqdm

from mlebench.utils import get_logger

logger = get_logger(__name__)


def prepare(raw: Path, public: Path, private: Path):
    DEV = False

    if DEV:
        batch_cutoff = 66  # 66 instead of 660 when in dev mode
    else:
        batch_cutoff = None

    logger.info("Loading raw metadata")
    old_train = pd.read_parquet(raw / "train_meta.parquet")

    # this has batch_id and event_id, we will do a test-train split based on batch_id
    # each batch id is equally sized so we can proceed with a simple split
    batch_ids = old_train["batch_id"].unique()[:batch_cutoff]

    logger.info("Splitting batches into train and test")
    train_batch_ids, test_batch_ids = train_test_split(batch_ids, test_size=0.1, random_state=0)

    # new column tracking the split
    old_train["split"] = None
    old_train.loc[old_train["batch_id"].isin(train_batch_ids), "split"] = "train"
    old_train.loc[old_train["batch_id"].isin(test_batch_ids), "split"] = "test"

    new_train = (
        old_train[old_train["split"] == "train"]
        .drop(columns=["split"])
        .reset_index(drop=True)
        .copy()
    )
    new_test = (
        old_train[old_train["split"] == "test"]
        .drop(columns=["split"])
        .reset_index(drop=True)
        .copy()
    )

    logger.info("Creating label-less test and sample submission")
    new_test_without_labels = new_test.drop(columns=["azimuth", "zenith"])

    # match sample submission format
    new_test = new_test[["event_id", "azimuth", "zenith"]]

    # copy the format as the private test and fill dummy values like kaggle.com
    sample_submission = new_test.copy()
    sample_submission["azimuth"] = 1
    sample_submission["zenith"] = 1

    logger.info("Saving files")
    # save the prepared tables
    new_train.to_parquet(public / "train_meta.parquet", index=False, engine="fastparquet")
    new_test_without_labels.to_parquet(
        public / "test_meta.parquet", index=False, engine="fastparquet"
    )
    sample_submission.to_csv(public / "sample_submission.csv", index=False)
    new_test.to_csv(private / "test.csv", index=False)

    logger.info("Copying remaining files")

    # sensor_geometry can be copied as is
    shutil.copy(raw / "sensor_geometry.csv", public / "sensor_geometry.csv")

    # copy the raw train files to train and test folders respectively
    train_batch_ids = set(train_batch_ids)
    train_dest = public / "train"
    train_dest.mkdir(exist_ok=True, parents=True)
    test_batch_ids = set(test_batch_ids)
    test_dest = public / "test"
    test_dest.mkdir(exist_ok=True, parents=True)
    for batch_file in tqdm(
        sorted((raw / "train").glob("*.parquet")), desc="Copying batch parquet files"
    ):
        batch_id = int(
            batch_file.stem.split("_")[-1]
        )  # i.e. go from e.g. 'train_000.parquet' to '000' to 0
        if batch_id in train_batch_ids:
            shutil.copy(batch_file, train_dest / batch_file.name)
        elif batch_id in test_batch_ids:
            shutil.copy(batch_file, test_dest / batch_file.name)

    logger.info("Running checks")
    # Asserts
    assert len(list(public.glob("train/*.parquet"))) == len(
        train_batch_ids
    ), "Not all train batches copied"
    assert len(list(public.glob("test/*.parquet"))) == len(
        test_batch_ids
    ), "Not all test batches copied"
    assert len(train_batch_ids) + len(test_batch_ids) == len(
        batch_ids
    ), "Something went wrong with splitting the batches"

    assert len(new_train) + len(new_test) == len(
        old_train[old_train["split"].notnull()]
    ), "Expected train + test to equal the original data"
    assert len(sample_submission) == len(
        new_test
    ), "Length mismatch between private test and sample submission"

    assert sample_submission.columns.equals(
        new_test.columns
    ), f"Column mismatch between sample_submission and private test"
    assert new_train.columns.equals(
        old_train.drop(columns=["split"]).columns
    ), f"Unexpected columns in train, expected {old_train.columns}, got {new_train.columns}"
    assert new_test_without_labels.columns.equals(
        old_train.drop(columns=["azimuth", "zenith", "split"]).columns
    ), f"Unexpected columns in test, expected {old_train.drop(columns=['azimuth', 'zenith']).columns}, got {new_test_without_labels.columns}"

    assert (
        len(set(new_train["event_id"]).intersection(set(new_test["event_id"]))) == 0
    ), "Event ids overlap between train and test"
    assert set(new_test["event_id"]) == set(
        sample_submission["event_id"]
    ), "Event ids mismatch between test and sample submission"
    logger.info("Done.")



import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

from mlebench.utils import read_csv


def _create_split_files(
    train_df: pd.DataFrame,
    test_df_with_solutions: pd.DataFrame,
    public_dir: Path,
    private_dir: Path,
):
    """
    Helper function to create the standard file structure for a given data split.

    This function encapsulates the logic for generating:
    - public/train.csv
    - public/test.csv (unlabeled)
    - public/sample_submission_null.csv
    - private/test.csv (labeled, ground truth)
    - private/gold_submission.csv
    """
    # Create directories if they don't exist
    public_dir.mkdir(exist_ok=True)
    private_dir.mkdir(exist_ok=True)

    # Save the training data
    train_df.to_csv(public_dir / "train.csv", index=False)

    # Save the full test set with solutions to the private directory
    test_df_with_solutions.to_csv(private_dir / "test.csv", index=False)

    # Create the gold submission from the private test set
    gold_submission = test_df_with_solutions[["Insult", "Date", "Comment"]].copy()
    gold_submission.to_csv(private_dir / "gold_submission.csv", index=False)

    # Create the public test set by dropping the label
    public_test = gold_submission.drop(columns=["Insult"]).copy()
    public_test.to_csv(public_dir / "test.csv", index=False)

    # Create a sample submission with null labels
    sample_submission = gold_submission.copy()
    sample_submission["Insult"] = 0
    sample_submission.to_csv(public_dir / "sample_submission_null.csv", index=False)


def prepare(raw: Path, public: Path, private: Path):
    # Load the original, pre-split data from the raw directory
    original_train_df = read_csv(raw / "train.csv")
    original_test_df = read_csv(raw / "test_with_solutions.csv")

    # --- Part 1: Generate the original public/private split ---
    # This block uses the original data to create the competition's primary
    # train/test split, ensuring the output is identical to the original script.
    _create_split_files(original_train_df, original_test_df, public, private)

    # --- Part 2: Generate the new validation split ---
    # This block creates a new split for local validation. It takes the original
    # training data and splits it again, creating a new, smaller training set
    # and a validation set. The outputs are saved to parallel directories.

    # Define paths for the new validation set directories
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Split the original training data to create a new train and validation set.
    # The size of the validation set is chosen to be the same as the original test set.
    # We stratify on the 'Insult' column to maintain label distribution.
    train_val_df, test_val_df = train_test_split(
        original_train_df,
        test_size=len(original_test_df),
        random_state=42,
        stratify=original_train_df["Insult"],
    )

    # Use the same helper function to create the files for the validation split,
    # ensuring an identical directory structure and naming convention.
    _create_split_files(train_val_df, test_val_df, public_val, private_val)

    # --- Final Checks ---
    # checks for the original split
    public_train = read_csv(public / "train.csv")
    public_test = read_csv(public / "test.csv")
    private_test = read_csv(private / "test.csv")
    sample_submission = read_csv(public / "sample_submission_null.csv")
    gold_submission = read_csv(private / "gold_submission.csv")

    assert public_train.columns.to_list() == [
        "Insult",
        "Date",
        "Comment",
    ], "Train columns should be Insult, Date, Comment"
    assert public_test.columns.to_list() == [
        "Date",
        "Comment",
    ], "Test columns should be Date, Comment"
    assert sample_submission.columns.to_list() == [
        "Insult",
        "Date",
        "Comment",
    ], "Sample submission columns should be Insult, Date, Comment"
    assert gold_submission.columns.to_list() == [
        "Insult",
        "Date",
        "Comment",
    ], "Gold submission columns should be Insult, Date, Comment"
    assert private_test.columns.to_list() == [
        "Insult",
        "Date",
        "Comment",
        "Usage",
    ], "Private test columns should be Insult, Date, Comment, Usage"

    assert set(public_train["Comment"]).isdisjoint(
        set(public_test["Comment"])
    ), "None of the test comments should be in the train comments"
    assert public_test.equals(
        private_test.drop(columns=["Insult", "Usage"])
    ), "Public test should be identical to private test, modulo the Insult and Usage columns"
    assert set(public_test["Comment"]) == set(
        sample_submission["Comment"]
    ), "Public test and sample submission should have the same Comments"
    assert set(public_test["Comment"]) == set(
        gold_submission["Comment"]
    ), "Public test and gold submission should have the same Comments"


import shutil
from pathlib import Path
import pandas as pd

from sklearn.model_selection import train_test_split
from tqdm.auto import tqdm

from mlebench.utils import compress, extract, read_csv


def _process_and_save_split(
    train_df: pd.DataFrame,
    test_df: pd.DataFrame,
    public_dir: Path,
    private_dir: Path,
    raw_image_dir: Path,
    original_df: pd.DataFrame,
):
    """
    Helper function to process a single train/test split.
    It saves all required CSVs, copies and compresses images, and runs assertions.
    """
    public_dir.mkdir(parents=True, exist_ok=True)
    private_dir.mkdir(parents=True, exist_ok=True)
    
    # Save the dataframes
    train_df.to_csv(public_dir / "train.csv", index=False)
    test_df.to_csv(private_dir / "test.csv", index=False)

    sample_submission = test_df.copy()
    sample_submission["has_cactus"] = 0.5
    sample_submission.to_csv(public_dir / "sample_submission.csv", index=False)

    # Prepare temporary directories for images
    temp_train_img_dir = public_dir / "train"
    temp_test_img_dir = public_dir / "test"
    temp_train_img_dir.mkdir(parents=True, exist_ok=True)
    temp_test_img_dir.mkdir(parents=True, exist_ok=True)

    # Copy images for the current split
    for image_id in tqdm(train_df["id"], desc=f"Copying train images to {public_dir.name}", total=len(train_df)):
        shutil.copy(raw_image_dir / image_id, temp_train_img_dir / image_id)

    for image_id in tqdm(test_df["id"], desc=f"Copying test images to {public_dir.name}", total=len(test_df)):
        shutil.copy(raw_image_dir / image_id, temp_test_img_dir / image_id)

    # and then recompress
    compress(temp_train_img_dir, public_dir / "train.zip")
    compress(temp_test_img_dir, public_dir / "test.zip")

    # and cleanup temporary image directories
    shutil.rmtree(temp_train_img_dir)
    shutil.rmtree(temp_test_img_dir)

    # checks for the current split
    assert (public_dir / "train.zip").exists(), f"{public_dir.name}/train.zip should exist"
    assert (public_dir / "test.zip").exists(), f"{public_dir.name}/test.zip should exist"

    assert len(train_df) + len(test_df) == len(
        original_df
    ), "The lengths of the splits should add up to the original"
    assert len(train_df) > len(test_df), "The train set should be larger than the test set"
    assert len(test_df) == len(
        sample_submission
    ), "The test set should match the sample submission"

    assert (
        train_df.columns.tolist()
        == test_df.columns.tolist()
        == original_df.columns.tolist()
        == sample_submission.columns.tolist()
    ), "All dataframes should have the same columns, i.e. ['id', 'has_cactus']"

    assert set(train_df["id"]).isdisjoint(test_df["id"]), "Train and test ids should not overlap"
    assert set(test_df["id"]) == set(
        sample_submission["id"]
    ), "Test and sample_submission ids should match"

    assert train_df["id"].nunique() == len(train_df), "There should be no duplicate ids in train"
    assert test_df["id"].nunique() == len(test_df), "There should be no duplicate ids in test"


def prepare(raw: Path, public: Path, private: Path):

    old_train = read_csv(raw / "train.csv")
    
    # need to split the train.zip into train.zip and test.zip; to do so need to extract first
    extract(raw / "train.zip", raw)
    raw_image_dir = raw / "train"

    # --- 1. Original Split: Create main train and test sets ---
    # This block creates the original public/private outputs, which must remain unchanged.
    
    # 4000 / (4000 + 17500) -> test_size is ~0.19
    new_train, new_test = train_test_split(old_train, test_size=0.19, random_state=0)
    
    _process_and_save_split(
        train_df=new_train,
        test_df=new_test,
        public_dir=public,
        private_dir=private,
        raw_image_dir=raw_image_dir,
        original_df=old_train,
    )
    
    # --- 2. New Validation Split: Create a validation set from the main train set ---
    # This block creates the new public_val/private_val outputs.
    
    # Define paths for the new validation split artifacts
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # To make the new validation set (test_val) have the same size as the original
    # test set (new_test), we calculate the required test_size proportion.
    test_size_for_val_split = len(new_test) / len(new_train)
    
    # Split the main training set (new_train) into a smaller training set (train_val)
    # and a validation set (test_val), using the same random_state for consistency.
    train_val, test_val = train_test_split(new_train, test_size=test_size_for_val_split, random_state=0)

    _process_and_save_split(
        train_df=train_val,
        test_df=test_val,
        public_dir=public_val,
        private_dir=private_val,
        raw_image_dir=raw_image_dir,
        original_df=new_train,
    )

    # --- Final Cleanup ---
    # and cleanup the extracted raw image directory, which is no longer needed
    shutil.rmtree(raw_image_dir)
    assert not (raw / "train").exists(), "raw/train/ should not exist"


import json
import shutil
from pathlib import Path
from typing import List, Dict

import pandas as pd
import py7zr
from sklearn.model_selection import train_test_split

from mlebench.utils import extract


def _create_split_files(
    train_set: List[Dict],
    test_set: List[Dict],
    test_fields: List[str],
    public_dir: Path,
    private_dir: Path,
):
    """
    Helper function to generate all files for a given train/test split.

    This function creates:
    - public/{train.json.7z, test.json.7z, sample_submission.csv.7z}
    - private/{test.csv, sample_submission.csv}
    """
    # Ensure output directories exist
    public_dir.mkdir(exist_ok=True, parents=True)
    private_dir.mkdir(exist_ok=True, parents=True)

    # Create test set without labels
    test_set_without_labels = []
    for sample in test_set:
        sample_copy = {key: sample[key] for key in test_fields}
        test_set_without_labels.append(sample_copy)

    # Write new train and test splits to a temporary directory
    tmp_data_dir = private_dir / "tmp_data"
    tmp_data_dir.mkdir(exist_ok=True)
    try:
        with open(tmp_data_dir / "train.json", "w") as f:
            json.dump(train_set, f)
        with open(tmp_data_dir / "test.json", "w") as f:
            json.dump(test_set_without_labels, f)

        # Compress the JSON files into the public directory
        with py7zr.SevenZipFile(public_dir / "train.json.7z", "w") as archive:
            archive.write(
                tmp_data_dir / "train.json",
                arcname="train.json",
            )
        with py7zr.SevenZipFile(public_dir / "test.json.7z", "w") as archive:
            archive.write(
                tmp_data_dir / "test.json",
                arcname="test.json",
            )

        # Make answers as csv from the labeled test set
        answer_rows = [
            {"id": sample["id"], "is_iceberg": int(sample["is_iceberg"])}
            for sample in test_set
        ]
        answers = pd.DataFrame(answer_rows)
        answers.to_csv(private_dir / "test.csv", index=False)

        # Make a sample submission file
        sample_submission = answers.copy()
        sample_submission["is_iceberg"] = 0.5
        sample_submission.to_csv(private_dir / "sample_submission.csv", index=False)
        with py7zr.SevenZipFile(public_dir / "sample_submission.csv.7z", "w") as archive:
            archive.write(
                private_dir / "sample_submission.csv",
                arcname="sample_submission.csv",
            )
    finally:
        # Ensure temporary files are removed
        shutil.rmtree(tmp_data_dir)

    # --- Final checks for this split ---
    assert len(test_set) == len(
        test_set_without_labels
    ), f"Expected test_set ({len(test_set)}) to have the same length as test_set_without_labels ({len(test_set_without_labels)})"
    assert len(answers) == len(
        test_set
    ), f"Expected answers ({len(answers)}) to have the same length as test_set ({len(test_set)})"
    assert len(sample_submission) == len(
        test_set
    ), f"Expected sample_submission ({len(sample_submission)}) to have the same length as test_set ({len(test_set)})"
    assert set(answers.columns) == set(
        ["id", "is_iceberg"]
    ), "Answers must have 'id' and 'is_iceberg' columns"
    assert set(sample_submission.columns) == set(
        ["id", "is_iceberg"]
    ), "Sample submission must have 'id' and 'is_iceberg' columns"
    train_ids = set([sample["id"] for sample in train_set])
    test_ids = set([sample["id"] for sample in test_set])
    assert train_ids.isdisjoint(test_ids), "Train and test ids should not overlap"


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Also creates a secondary validation split (public_val, private_val) for model development.
    """
    extract(raw / "train.json.7z", raw)
    extract(raw / "test.json.7z", raw)
    old_train = json.load((raw / "data/processed/train.json").open())
    old_test = json.load((raw / "data/processed/test.json").open())

    all_fields = list([key for key in old_train[0].keys()])
    assert all(
        set(all_fields) == set([key for key in sample.keys()]) for sample in old_train
    ), "Inconsistent fields in train set"
    test_fields = list([key for key in old_test[0].keys()])
    assert all(
        set(test_fields) == set([key for key in sample.keys()]) for sample in old_test
    ), "Inconsistent fields in test set"

    # --- First Split: Create the main train/test sets for the competition ---
    # Old ratio is Train set - 1,604 samples; Test set - 8,424 samples (~84% ratio)
    # We do a 20% ratio to avoid removing too many samples from train
    new_train, new_test = train_test_split(old_train, test_size=0.2, random_state=0)

    # Generate the original public and private directory files
    _create_split_files(
        train_set=new_train,
        test_set=new_test,
        test_fields=test_fields,
        public_dir=public,
        private_dir=private,
    )

    # Check that the total number of samples is conserved in the first split
    assert len(new_train) + len(new_test) == len(
        old_train
    ), f"Expected {len(old_train)} total samples in new_train ({len(new_train)}) and new_test ({len(new_test)})"

    # --- Second Split: Create a validation set from the main training set ---
    # Define paths for the new validation split
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # The goal is for the new validation set (test_val) to have the same size as the original test set (new_test).
    # test_size = len(new_test) / len(new_train) = (0.2 * N) / (0.8 * N) = 0.25
    test_size_for_val_split = len(new_test) / len(new_train)
    train_val, test_val = train_test_split(
        new_train, test_size=test_size_for_val_split, random_state=0
    )

    # Generate the validation public_val and private_val directory files
    _create_split_files(
        train_set=train_val,
        test_set=test_val,
        test_fields=test_fields,
        public_dir=public_val,
        private_dir=private_val,
    )


import os
from pathlib import Path
from typing import Any
import pandas as pd

# This is a generic LLM-based grader for open-ended tasks.
# It reads 'rubric.md' from the task directory and evaluates the submission.

try:
    from dsat.services.llm import LLMService
    from dsat.config import LLMConfig
except ImportError:
    # Fallback for when running outside of dsat package context
    import sys
    sys.path.append(str(Path(__file__).resolve().parent.parent.parent.parent))
    from dsat.services.llm import LLMService
    from dsat.config import LLMConfig

class Report:
    def __init__(self, score, feedback):
        self.score = score
        self.feedback = feedback
        # Standard fields expected by the framework
        self.is_lower_better = False
        self.submission_exists = True
        self.valid_submission = True
        self.gold_medal = score >= 0.9
        self.silver_medal = score >= 0.7
        self.bronze_medal = score >= 0.5
        self.above_median = score >= 0.5
        self.submission_path = ""
        self.competition_id = "open_ended_task"

def grade(submission_path: Path, competition: Any) -> Report:
    """
    Grades the submission using an LLM Judge against rubric.md.
    """
    # 1. Load the Rubric
    task_dir = competition.raw_dir.parent
    rubric_path = task_dir / "rubric.md"
    
    if not rubric_path.exists():
        # Fallback if no rubric exists
        print(f"Warning: Rubric not found at {rubric_path}. Returning default score.")
        return Report(0.5, "No grading rubric defined.")
        
    rubric_content = rubric_path.read_text(encoding="utf-8")
    
    # 2. Load the Submission Content (Preview)
    # Since it's open-ended, the 'submission_path' might be a CSV, code, or just a marker.
    # We'll try to peek at the output artifacts if possible, or assume the agent's recent work 
    # is what we are grading. Ideally, AIDE produces a submission file.
    
    submission_content = "No submission content readable."
    if submission_path.exists():
        try:
            if submission_path.suffix == '.csv':
                df = pd.read_csv(submission_path)
                submission_content = f"CSV Submission Preview:\n{df.head().to_markdown()}"
            else:
                submission_content = submission_path.read_text(encoding="utf-8")[:2000]
        except Exception as e:
            submission_content = f"Error reading submission: {e}"

    # 3. Setup LLM for Judging
    # Note: In a real run, we might want to inject the API key securely.
    # Here we assume environment variables are set (which they are in DSATRunner).
    try:
        api_key = os.getenv("API_KEY", "EMPTY")
        base_url = os.getenv("API_BASE", "https://api.openai.com/v1")
        model = os.getenv("LLM_MODEL", "gpt-4o")
        
        llm = LLMService(LLMConfig(api_key=api_key, api_base=base_url, model=model))
        
        prompt = f"""You are an impartial Judge. Evaluate the following submission against the provided Rubric.

# RUBRIC
{rubric_content}

# SUBMISSION CONTENT
{submission_content}

# INSTRUCTION
Assess the submission. 
Output ONLY a float number between 0.0 and 1.0 on the first line.
On subsequent lines, provide brief feedback.
"""
        # Synchronous call wrapper or direct call if possible. 
        # Since grade() is synchronous in standard mlebench, we need a way to run async code.
        import asyncio
        response = asyncio.run(llm.achat([{"role": "user", "content": prompt}]))
        
        lines = response.strip().split('\n')
        try:
            score = float(lines[0].strip())
        except ValueError:
            # Fallback if LLM is chatty
            import re
            match = re.search(r"(\d+(\.\d+)?)", lines[0])
            score = float(match.group(1)) if match else 0.5
            
        feedback = "\n".join(lines[1:])
        return Report(score, feedback)

    except Exception as e:
        print(f"LLM Judging failed: {e}")
        return Report(0.0, f"Judging failed: {e}")



import shutil
from pathlib import Path

from sklearn.model_selection import train_test_split

from mlebench.competitions.utils import df_to_one_hot
from mlebench.utils import read_csv

from .dogs import DOGS_LIST


def to_one_hot(df):
    return df_to_one_hot(df, id_column="id", target_column="breed", classes=DOGS_LIST)


def prepare(raw: Path, public: Path, private: Path):

    # Create train, test from train split
    old_train = read_csv(raw / "labels.csv")
    new_train, new_test = train_test_split(old_train, test_size=0.1, random_state=0)
    # one hot the private test because we will one-hot the submission, as per kaggle.com
    new_test = to_one_hot(new_test)

    # Copy over files
    new_train.to_csv(public / "labels.csv", index=False)
    new_test.to_csv(private / "test.csv", index=False)

    (public / "train").mkdir(exist_ok=True)

    for file_id in new_train["id"]:
        shutil.copyfile(
            src=raw / "train" / f"{file_id}.jpg",
            dst=public / "train" / f"{file_id}.jpg",
        )

    (public / "test").mkdir(exist_ok=True)

    for file_id in new_test["id"]:
        shutil.copyfile(
            src=raw / "train" / f"{file_id}.jpg",
            dst=public / "test" / f"{file_id}.jpg",
        )

    # Check integrity of the files copied
    assert len(list(public.glob("train/*.jpg"))) == len(new_train)
    assert len(list(public.glob("test/*.jpg"))) == len(new_test)

    # Create a sample submission file
    submission_df = new_test.copy()
    for col in submission_df.columns[1:]:
        submission_df[col] = submission_df[col].astype("float64")
    submission_df.iloc[:, 1:] = 1 / 120
    submission_df.to_csv(public / "sample_submission.csv", index=False)

    assert submission_df.shape == (len(new_test), 121)  # 1 id column + 120 breeds



import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split


def get_date(s: str) -> str:
    """Gets date from string in the format YYYY-MM-DD-X where `X` is an arbitrary string."""

    split = s.split("-")

    assert (
        len(split) >= 3
    ), f"Expected the string to have at least 3 parts separated by `-`. Got {len(split)} parts."

    year, month, day = split[:3]

    assert (
        isinstance(year, str) and year.isdigit()
    ), f"Expected the year to be a string of digits. Got {year} instead."

    assert (
        isinstance(month, str) and month.isdigit()
    ), f"Expected the month to be a string of digits. Got {month} instead."

    assert (
        isinstance(day, str) and day.isdigit()
    ), f"Expected the day to be a string of digits. Got {day} instead."

    date = f"{year}-{month.zfill(2)}-{day.zfill(2)}"

    return date


def _process_split(
    raw_data_path: Path,
    train_ids: list,
    test_ids: list,
    public_path: Path,
    private_path: Path,
) -> None:
    """
    Helper function to process a single data split.

    It populates the public and private directories with the provided train/test IDs,
    creating the necessary file structure and artifacts (like the sample submission).
    """
    # Clean and create output directories
    shutil.rmtree(public_path, ignore_errors=True)
    shutil.rmtree(private_path, ignore_errors=True)
    public_path.mkdir(parents=True)
    private_path.mkdir(parents=True)
    (public_path / "train").mkdir()
    (public_path / "test").mkdir()

    for train_id in train_ids:
        shutil.copytree(
            src=raw_data_path / "train" / train_id,
            dst=public_path / "train" / train_id,
        )

    for test_id in test_ids:
        shutil.copytree(
            src=raw_data_path / "train" / test_id,
            dst=public_path / "test" / test_id,
        )

    # Construct test set by concatenating all ground truth csvs for the test journeys
    dfs = []
    for fpath in sorted((public_path / "test").rglob("ground_truth.csv")):
        drive_id = fpath.parent.parent.name
        phone_id = fpath.parent.name

        assert (
            drive_id in test_ids
        ), f"Expected the drive {drive_id} to be one of the new test instances. Got {drive_id} instead."

        raw_df = pd.read_csv(fpath)
        df = raw_df.copy()
        df.loc[:, "tripId"] = f"{drive_id}-{phone_id}"
        df = df[["tripId", "UnixTimeMillis", "LatitudeDegrees", "LongitudeDegrees"]]
        dfs.append(df)

    new_test_labels = pd.concat(dfs, ignore_index=True)
    # The output filename is 'test.csv' to match the competition structure.
    new_test_labels.to_csv(private_path / "test.csv", index=False)

    for fpath in (public_path / "test").rglob("ground_truth.csv"):
        fpath.unlink()  # don't include ground truth in public test data

    shutil.copytree(
        src=raw_data_path / "metadata",
        dst=public_path / "metadata",
    )

    actual_journey_ids = set(["-".join(s.split("-")[:-1]) for s in new_test_labels["tripId"]])
    assert len(actual_journey_ids) == len(test_ids), (
        f"Expected the new test instances to have {len(test_ids)} unique trip IDs. Got "
        f"{len(new_test_labels['tripId'].unique())} unique trip IDs."
    )

    sample_submission = new_test_labels.copy()
    sample_submission.loc[:, "LatitudeDegrees"] = 37.904611315634504
    sample_submission.loc[:, "LongitudeDegrees"] = -86.48107806249548

    assert len(sample_submission) == len(new_test_labels), (
        f"Expected the sample submission to have the same number of instances as the new test "
        f"instances. Got {len(sample_submission)} instances in the sample submission and "
        f"{len(new_test_labels)} new test instances."
    )

    sample_submission.to_csv(public_path / "sample_submission.csv", index=False)

    assert sorted(list(public_path.glob("train/*"))) == sorted(
        set([public_path / "train" / drive_id for drive_id in train_ids])
    ), "Expected the public train directory to contain the new train instances."

    assert sorted(list(public_path.glob("test/*"))) == sorted(
        set([public_path / "test" / drive_id for drive_id in test_ids])
    ), "Expected the public test directory to contain the new test instances."

    assert (
        len(list((public_path / "test").rglob("ground_truth.csv"))) == 0
    ), "Expected the public test directory to not contain any ground truth files."

    assert len(list((public_path / "train").rglob("ground_truth.csv"))) >= len(train_ids), (
        "Expected the public train directory to contain at least one ground truth file per new "
        "train instance."
    )


def prepare(raw: Path, public: Path, private: Path) -> None:
    # --- Stage 1: Original Split (Train / Test) ---
    # This section creates the primary competition data in `public` and `private`.
    # Its logic and outputs are identical to the original script.

    old_train_ids = sorted([folder.name for folder in (raw / "train").glob("*") if folder.is_dir()])
    dates = sorted(set([get_date(s) for s in old_train_ids]))
    new_train_dates, new_test_dates = train_test_split(dates, test_size=0.1, random_state=0)

    assert (
        len(new_train_dates) >= 1
    ), "Expected the new train set to have at least one date. Got 0 dates."

    assert (
        len(new_test_dates) >= 1
    ), "Expected the new test set to have at least one date. Got 0 dates."

    new_train_ids = sorted([i for i in old_train_ids if get_date(i) in new_train_dates])
    new_test_ids = sorted([i for i in old_train_ids if get_date(i) in new_test_dates])

    assert len(set(new_train_ids).intersection(set(new_test_ids))) == 0, (
        f"Expected the new train and test instances to be disjoint. Got an intersection of "
        f"{set(new_train_ids).intersection(set(new_test_ids))}."
    )

    assert len(new_train_ids) + len(new_test_ids) == len(old_train_ids), (
        f"Expected the number of new train and test instances to sum up to the number of old train "
        f"instances. Got {len(new_train_ids)} new train instances and {len(new_test_ids)} new test "
        f"instances which sum to {len(new_train_ids) + len(new_test_ids)} instead of "
        f"{len(old_train_ids)}."
    )

    _process_split(
        raw_data_path=raw,
        train_ids=new_train_ids,
        test_ids=new_test_ids,
        public_path=public,
        private_path=private,
    )

    # --- Stage 2: Validation Split (Train_val / Test_val) ---
    # This section creates a new validation dataset in parallel directories.
    # It takes the `new_train_ids` from the first split and splits them again
    # using the exact same methodology.

    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # The input for this split is the training set from the *first* split.
    val_split_input_dates = sorted(set([get_date(s) for s in new_train_ids]))

    # We replicate the splitting logic and parameters to get a test set of a
    # similar proportional size to the original test set.
    train_val_dates, test_val_dates = train_test_split(
        val_split_input_dates, test_size=0.1, random_state=0
    )

    train_val_ids = sorted([i for i in new_train_ids if get_date(i) in train_val_dates])
    test_val_ids = sorted([i for i in new_train_ids if get_date(i) in test_val_dates])

    _process_split(
        raw_data_path=raw,
        train_ids=train_val_ids,
        test_ids=test_val_ids,
        public_path=public_val,
        private_path=private_val,
    )


# dsat/services/states/autokaggle_state.py

from pydantic import BaseModel, Field
from typing import List, Dict, Any, Optional

from dsat.models.formats import TaskContract


class AttemptMemory(BaseModel):
    """Stores the memory of a single attempt within a phase."""
    attempt_number: int
    plan: str
    code: str
    execution_output: str
    execution_error: Optional[str]
    validation_result: Dict[str, Any]
    review_score: float
    review_suggestion: str


class PhaseMemory(BaseModel):
    """Stores the history of all attempts for a single successful phase."""
    phase_goal: str
    attempts: List[AttemptMemory] = Field(default_factory=list)
    final_report: str = ""
    is_successful: bool = False
    # Keep track of artifacts produced *in this phase*. Paths are relative to the sandbox workdir.
    output_artifacts: Dict[str, str] = Field(default_factory=dict, description="Mapping of artifact filenames to their descriptions.")


class AutoKaggleState(BaseModel):
    """Manages the entire state of a dynamic AutoKaggle workflow run."""
    contract: TaskContract
    dynamic_phases: List[str] = Field(default_factory=list)
    phase_history: List[PhaseMemory] = Field(default_factory=list)
    io_instructions: str = Field(default="", description="Standardized I/O instructions for the agents.")
    full_task_description: str = Field(
        description="The original, complete task description including the data analysis report from DataAnalyzer."
    )
    # The central registry for all artifacts created during the run.
    # This provides state continuity between phases.
    # Maps artifact filename to a description.
    global_artifacts: Dict[str, str] = Field(default_factory=dict)


import shutil
from pathlib import Path

from sklearn.model_selection import train_test_split
from tqdm.auto import tqdm

from mlebench.utils import read_csv


def prepare(raw: Path, public: Path, private: Path):
    old_train = read_csv(raw / "train.csv")
    # drop image ce27d36c9147cc19.jpg: it appears twice and may occur across train and test when split
    old_train = old_train[old_train["image"] != "ce27d36c9147cc19.jpg"]

    new_train, new_test = train_test_split(old_train, test_size=0.1, random_state=0)
    sample_submission = new_test.copy()[["image", "hotel_id"]]
    sample_submission["hotel_id"] = "36363 53586 18807 64314 60181"

    # save public files
    new_train.to_csv(public / "train.csv", index=False)
    sample_submission.to_csv(public / "sample_submission.csv", index=False)

    # copy images from raw to prepared/public accordingly
    (public / "train_images").mkdir(exist_ok=True, parents=True)
    for image, chain in tqdm(
        zip(new_train["image"], new_train["chain"]), total=len(new_train), desc="Train images"
    ):
        chain = str(chain)
        (public / "train_images" / chain).mkdir(exist_ok=True, parents=True)
        shutil.copy(raw / "train_images" / chain / image, public / "train_images" / chain / image)

    (public / "test_images").mkdir(exist_ok=True, parents=True)
    for image, chain in tqdm(
        zip(new_test["image"], new_test["chain"]), total=len(new_test), desc="Test images"
    ):
        chain = str(chain)
        (public / "test_images").mkdir(exist_ok=True, parents=True)
        shutil.copy(raw / "train_images" / chain / image, public / "test_images" / image)

    # save private files
    new_test.to_csv(private / "test.csv", index=False)

    # checks
    assert len(new_train) + len(new_test) == len(
        old_train
    ), "Length of new_train and new_test should equal length of old_train"
    assert sample_submission.columns.to_list() == [
        "image",
        "hotel_id",
    ], "Sample submission columns should only be `image` and `hotel_id`"
    assert len(sample_submission) == len(new_test), "Sample submission length should match test set"
    for image, chain in zip(new_train["image"], new_train["chain"]):
        chain = str(chain)
        assert (
            public / "train_images" / chain / image
        ).exists(), f"Image {image} not found in train_images folder"
    for image in new_test["image"]:
        assert (
            public / "test_images" / image
        ).exists(), f"Image {image} not found in test_images folder"
    assert not set(new_train["image"]).intersection(
        set(new_test["image"])
    ), "Train and test ids overlap"



import json
import shutil
from pathlib import Path
from typing import List, Dict, Any

import pandas as pd
from sklearn.model_selection import train_test_split


def _create_split(
    data_to_split: List[Dict],
    test_size: float,
    test_fields: List[str],
    public_path: Path,
    private_path: Path,
    random_state: int,
) -> List[Dict]:
    """
    Helper function to perform a data split and create all required files.

    Args:
        data_to_split: The list of data samples to be split.
        test_size: The proportion of the dataset to allocate to the test split.
        test_fields: The list of fields to keep in the test set.
        public_path: The directory for public-facing files (train set, unlabeled test set).
        private_path: The directory for private files (test set labels).
        random_state: The seed for the random number generator.

    Returns:
        The training portion of the split, to be used for subsequent splits if needed.
    """
    # Create output directories if they don't exist
    public_path.mkdir(parents=True, exist_ok=True)
    private_path.mkdir(parents=True, exist_ok=True)

    # Create train, test from the provided data
    train_set, test_set = train_test_split(
        data_to_split, test_size=test_size, random_state=random_state
    )

    test_set_without_labels = []
    for sample in test_set:
        # Keep only the fields that should be in the test set
        sample_unlabeled = {key: sample[key] for key in test_fields}
        test_set_without_labels.append(sample_unlabeled)

    # Save the new train and test splits to the public directory
    with open(public_path / "train.json", "w") as f:
        json.dump(train_set, f, indent=4)
    with open(public_path / "test.json", "w") as f:
        json.dump(test_set_without_labels, f, indent=4)
    # Also save zipped versions
    shutil.make_archive(public_path / "train", "zip", public_path, "train.json")
    shutil.make_archive(public_path / "test", "zip", public_path, "test.json")

    # Create answers for the private directory
    answers_rows = []
    for sample in test_set:
        answers_rows.append(
            {
                "request_id": sample["request_id"],
                "requester_received_pizza": int(sample["requester_received_pizza"]),
            }
        )
    answers = pd.DataFrame(answers_rows)
    answers.to_csv(private_path / "test.csv", index=False)

    # Create sample submission for the public directory
    sample_submission = answers.copy()
    sample_submission["requester_received_pizza"] = 0
    sample_submission.to_csv(public_path / "sampleSubmission.csv", index=False)

    # Perform checks
    assert len(train_set) + len(test_set) == len(
        data_to_split
    ), f"Expected {len(data_to_split)} total samples, but got {len(train_set)} in train and {len(test_set)} in test"
    assert len(test_set) == len(
        test_set_without_labels
    ), "Test set and unlabeled test set must have the same length"
    assert len(answers) == len(test_set), "Answers must have the same length as the test set"
    assert len(sample_submission) == len(
        test_set
    ), "Sample submission must have the same length as the test set"
    assert set(answers.columns) == set(
        ["request_id", "requester_received_pizza"]
    ), "Answers must have 'request_id' and 'requester_received_pizza' columns"
    assert set(sample_submission.columns) == set(
        ["request_id", "requester_received_pizza"]
    ), "Sample submission must have 'request_id' and 'requester_received_pizza' columns"

    return train_set


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Then, it creates a second, parallel validation split from the first training set.
    """

    # Load data
    with open(raw / "train.json") as f:
        old_train = json.load(f)
    with open(raw / "test.json") as f:
        old_test = json.load(f)

    test_ratio = len(old_test) / (len(old_train) + len(old_test))

    all_fields = list([key for key in old_train[0].keys()])
    assert all(set(all_fields) == set([key for key in sample.keys()]) for sample in old_train)
    test_fields = list([key for key in old_test[0].keys()])
    assert all(set(test_fields) == set([key for key in sample.keys()]) for sample in old_test)

    # --- Original Split ---
    # This split creates the primary `public` and `private` competition data.
    # The returned `new_train` set will be used for the subsequent validation split.
    new_train = _create_split(
        data_to_split=old_train,
        test_size=test_ratio,
        test_fields=test_fields,
        public_path=public,
        private_path=private,
        random_state=0,
    )

    # --- New Validation Split ---
    # Define new directories for the validation set, parallel to the original ones.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # To keep the new test set (`test_val`) size consistent with the original test set,
    # we adjust the test ratio for the second split.
    # val_test_size = original_test_size / original_train_size
    val_test_size = test_ratio / (1.0 - test_ratio)

    # Create the validation split using the same logic, but on the `new_train` data
    # and saving to the new `_val` directories.
    _create_split(
        data_to_split=new_train,
        test_size=val_test_size,
        test_fields=test_fields,
        public_path=public_val,
        private_path=private_val,
        random_state=0,
    )


"""
Global configuration for DSLighting.

This module provides a central place to store configuration that can be shared
across the DSLighting package, such as parent directories for data and registry.
"""

from pathlib import Path
from typing import Optional
from threading import Lock

class GlobalConfig:
    """
    Thread-safe global configuration for DSLighting.

    This class stores configuration that can be shared across Agent instances,
    such as default parent directories for data and registry.

    Attributes:
        data_parent_dir: Default parent directory for competition data
        registry_parent_dir: Default parent directory for registry configurations
    """
    _instance = None
    _lock = Lock()

    def __new__(cls):
        """Ensure singleton pattern with thread safety."""
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
                    cls._instance._data_parent_dir = None
                    cls._instance._registry_parent_dir = None
        return cls._instance

    @property
    def data_parent_dir(self) -> Optional[Path]:
        """Get the default data parent directory."""
        return self._data_parent_dir

    @data_parent_dir.setter
    def data_parent_dir(self, value: Optional[str | Path]):
        """Set the default data parent directory."""
        self._data_parent_dir = Path(value) if value is not None else None

    @property
    def registry_parent_dir(self) -> Optional[Path]:
        """Get the default registry parent directory."""
        return self._registry_parent_dir

    @registry_parent_dir.setter
    def registry_parent_dir(self, value: Optional[str | Path]):
        """Set the default registry parent directory."""
        self._registry_parent_dir = Path(value) if value is not None else None

    def set_parent_dirs(
        self,
        data_parent_dir: Optional[str | Path] = None,
        registry_parent_dir: Optional[str | Path] = None
    ):
        """
        Set default parent directories for data and registry.

        Args:
            data_parent_dir: Default parent directory for competition data.
                           Example: "/path/to/data/competitions"
            registry_parent_dir: Default parent directory for registry configurations.
                                Example: "/path/to/registry"

        Examples:
            >>> config = GlobalConfig()
            >>> config.set_parent_dirs(
            ...     data_parent_dir="/path/to/data/competitions",
            ...     registry_parent_dir="/path/to/registry"
            ... )
        """
        if data_parent_dir is not None:
            self.data_parent_dir = data_parent_dir
        if registry_parent_dir is not None:
            self.registry_parent_dir = registry_parent_dir

    def get_task_paths(self, task_id: str) -> tuple[Optional[Path], Optional[Path]]:
        """
        Get full paths for a task using configured parent directories.

        Args:
            task_id: Task/competition identifier (e.g., "bike-sharing-demand")

        Returns:
            Tuple of (data_dir, registry_dir) or (None, None) if not configured

        Examples:
            >>> config = GlobalConfig()
            >>> config.set_parent_dirs(
            ...     data_parent_dir="/path/to/data/competitions",
            ...     registry_parent_dir="/path/to/registry"
            ... )
            >>> data_dir, registry_dir = config.get_task_paths("bike-sharing-demand")
            >>> print(data_dir)
            /path/to/data/competitions/bike-sharing-demand
            >>> print(registry_dir)
            /path/to/registry/bike-sharing-demand
        """
        data_dir = None
        registry_dir = None

        if self._data_parent_dir is not None:
            data_dir = self._data_parent_dir / task_id

        if self._registry_parent_dir is not None:
            registry_dir = self._registry_parent_dir / task_id

        return data_dir, registry_dir

    def reset(self):
        """Reset all configuration to default values (None)."""
        self._data_parent_dir = None
        self._registry_parent_dir = None

    def __repr__(self) -> str:
        """Return string representation of current configuration."""
        return (
            f"GlobalConfig("
            f"data_parent_dir={self._data_parent_dir}, "
            f"registry_parent_dir={self._registry_parent_dir}"
            f")"
        )


# Global instance
_global_config = GlobalConfig()


def get_global_config() -> GlobalConfig:
    """
    Get the global configuration instance.

    Returns:
        GlobalConfig singleton instance

    Examples:
        >>> from dslighting.core.global_config import get_global_config
        >>> config = get_global_config()
        >>> config.set_parent_dirs(
        ...     data_parent_dir="/path/to/data",
        ...     registry_parent_dir="/path/to/registry"
        ... )
    """
    return _global_config



from bisect import bisect
from typing import Any

import pandas as pd

from mlebench.grade_helpers import InvalidSubmissionError


def kendall_tau(ground_truth: list[list[Any]], predictions: list[list[Any]]) -> float:
    """
    Computes the Kendall Tau correlation between
    - `ground_truth`: a list of lists, where each sublist contains the ground truth order of items
    - `predictions`: a list of lists, where each sublist contains the predicted order of items

    See: https://www.kaggle.com/code/ryanholbrook/competition-metric-kendall-tau-correlation
    """

    # Actually O(N^2), but fast in practice for our data
    def count_inversions(a):
        inversions = 0
        sorted_so_far = []
        for i, u in enumerate(a):  # O(N)
            j = bisect(sorted_so_far, u)  # O(log N)
            inversions += i - j
            sorted_so_far.insert(j, u)  # O(N)
        return inversions

    total_inversions = 0  # total inversions in predicted ranks across all instances
    total_2max = 0  # maximum possible inversions across all instances
    for gt, pred in zip(ground_truth, predictions):
        ranks = [gt.index(x) for x in pred]  # rank predicted order in terms of ground truth
        total_inversions += count_inversions(ranks)
        n = len(gt)
        total_2max += n * (n - 1)
    return 1 - 4 * total_inversions / total_2max


def prepare_for_metric(submission: pd.DataFrame, answers: pd.DataFrame) -> dict:
    if len(submission) != len(answers):
        raise InvalidSubmissionError("Submission and answers must have the same length")
    if "id" not in submission.columns or "cell_order" not in submission.columns:
        raise InvalidSubmissionError("Submission must have 'id' and 'cell_order' columns")
    if set(submission["id"]) != set(answers["id"]):
        raise InvalidSubmissionError("Submission and answers must have the same ids")
    # "cell_order" columns must be strings of space-separated cell ids: "54c7cab3 fe66203e 7844d5f8 ..."
    submission["cell_order"] = submission["cell_order"].astype(str)
    answers["cell_order"] = answers["cell_order"].astype(str)

    # Sort by id to ensure we're comparing the same rows
    submission = submission.sort_values("id")
    answers = answers.sort_values("id")

    ground_truth, predictions = [], []
    target_column = "cell_order"
    for (_, row_gt), (_, row_pred) in zip(submission.iterrows(), answers.iterrows()):
        ground_truth_items = row_gt[target_column].split()
        prediction_items = row_pred[target_column].split()
        if len(ground_truth_items) != len(prediction_items):
            raise InvalidSubmissionError("Ground truth and predictions must have the same length")
        ground_truth.append(ground_truth_items)
        predictions.append(prediction_items)
    return {
        "ground_truth": ground_truth,
        "predictions": predictions,
    }


def grade(submission: pd.DataFrame, answers: pd.DataFrame) -> float:
    kendall_tau_inputs = prepare_for_metric(submission, answers)
    return kendall_tau(**kendall_tau_inputs)



import json
import random
import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import get_logger

logger = get_logger(__name__)


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the raw data into public and private datasets with appropriate test/train splits.

    `train/metadata.json` is the "table of contents" for our data, with the following structure:
    (More details at https://www.kaggle.com/competitions/herbarium-2020-fgvc7/data)
    ```
    {
        "annotations" : [annotation],
        "categories" : [category],
        "images" : [image],
        "info" : info,
        "licenses" : [license],
        "regions" : [region]
    }
    ```
    - `images` and `annotations` are both N-length lists corresponding to the N samples.
        We'll need to split each of these lists into train and test.
    - The other fields are dataset-wide metadata that we don't need to touch.

    - test/metadata.json is the same structure as train/metadata.json, but without "annotations", "categories", "regions"

    Other notes:
    - train/test splits need to occur per category (each category should be in both train and test).
    - The `test/images` and `train/images` folders have nested subdirs to make it easier to browse
        - `train/images` is structured as `{category_id[:3]}/{category_id[3:]}/{image_id}.jpg`
        - `test/images` is structured as `{image_idx[:3]}/{image_idx}.jpg` (to not reveal the category)
    - When we create the new splits, we re-assign image indices so that we don't give away labels based on the index
        - train images are indexed within their own category
        - test images follow a flat index after shuffling the categories
    """

    dev_mode = False
    dev_count = 2  # Copy over n images per category when in dev mode

    # Create train, test from train split
    json_path = raw / "nybg2020/train/metadata.json"
    with open(json_path, "r", encoding="latin-1") as f:  # utf-8 fails
        old_train_metadata = json.load(f)

    # Organize data by category so that we can split per-category later
    annotations_images_by_category = {}  # We'll collect both `annotations` and `images` here
    for annotation, image in list(
        zip(old_train_metadata["annotations"], old_train_metadata["images"])
    ):
        assert (
            annotation["image_id"] == image["id"]
        ), f"Mismatching image_id in annotation and image: {annotation['image_id']} vs {image['id']}"
        category_id = annotation["category_id"]
        if category_id not in annotations_images_by_category:
            annotations_images_by_category[category_id] = []
        annotations_images_by_category[category_id].append(
            {
                "annotation": annotation,
                "image": image,
            }
        )

    # Split train/test
    train_sample_count = 0  # Useful for tqdm later
    train_annotations_images_by_category = {}
    test_annotations_images_by_category = {}

    for category_id, annotations_images in tqdm(
        annotations_images_by_category.items(), desc="Assigning train/test splits"
    ):
        # Create split by "category" (class): Each category needs to be in both train and test (80:20) as per original ratio
        test_size = 0.2
        n_samples = len(annotations_images)
        if n_samples == 1:
            # If only one sample, put it in train
            train_annotations_images = annotations_images
            test_annotations_images = []
        elif n_samples < 5:  # Minimum 5 samples to ensure at least 1 in test
            # Ensure at least 1 sample in test
            test_size = max(1, int(n_samples * test_size))
            train_annotations_images = annotations_images[:-test_size]
            test_annotations_images = annotations_images[-test_size:]
        else:
            train_annotations_images, test_annotations_images = train_test_split(
                annotations_images, test_size=test_size, random_state=0
            )

        train_annotations_images_by_category[category_id] = train_annotations_images
        test_annotations_images_by_category[category_id] = test_annotations_images
        train_sample_count += len(train_annotations_images)

    # Add to train set
    new_train_metadata = (
        old_train_metadata.copy()
    )  # Keep 'categories', 'info', 'licenses', 'regions'
    new_train_metadata.update(
        {
            "annotations": [],
            "images": [],
        }
    )
    with tqdm(
        desc="Creating new train dataset",
        total=train_sample_count,
    ) as pbar:
        for category_id, annotations_images in train_annotations_images_by_category.items():
            # Create a nested directory from category_id, e.g. 15504 -> "155/04" or 3 -> "000/03"
            category_subdir = f"{category_id // 100:03d}/{category_id % 100:02d}"
            (public / "nybg2020/train/images" / category_subdir).mkdir(exist_ok=True, parents=True)
            for idx, annotation_image in enumerate(annotations_images):
                new_annotation = annotation_image["annotation"].copy()
                new_train_metadata["annotations"].append(new_annotation)

                new_image = annotation_image["image"].copy()
                new_train_metadata["images"].append(new_image)

                # Copy file from raw to public
                if (
                    not dev_mode or idx < dev_count
                ):  # if dev_mode, only copy the first dev_count images
                    src_path = raw / "nybg2020/train" / annotation_image["image"]["file_name"]
                    dst_path = public / "nybg2020/train" / annotation_image["image"]["file_name"]
                    shutil.copyfile(src=src_path, dst=dst_path)

                pbar.update(1)

    with open(public / "nybg2020/train/metadata.json", "w") as f:
        json.dump(new_train_metadata, f, indent=4, sort_keys=True)

    if not dev_mode:
        assert len(list((public / "nybg2020/train/images").glob("**/*.jpg"))) == len(
            new_train_metadata["images"]
        ), f"Mismatching number of images in train_images, got {len(list((public / 'nybg2020/train/images').glob('**/*.jpg')))}"
        assert len(new_train_metadata["annotations"]) == len(
            new_train_metadata["images"]
        ), f"Mismatching number of annotations in train_metadata, got {len(new_train_metadata['annotations'])}"

    # Add to test set
    new_test_metadata = old_train_metadata.copy()
    del new_test_metadata["categories"]
    del new_test_metadata["regions"]
    new_test_metadata.update(
        {
            "annotations": [],
            "images": [],
        }
    )
    # Flatten and shuffle test set so that we don't have all the same categories in a row
    test_annotations_images = [
        item for sublist in test_annotations_images_by_category.values() for item in sublist
    ]
    random.Random(0).shuffle(test_annotations_images)
    for idx, annotation_image in tqdm(
        enumerate(test_annotations_images),
        desc="Creating new test dataset",
        total=len(test_annotations_images),
    ):

        # Make new image id, for test set this is just the index
        new_image_id = str(idx)
        # Make new filename from image id e.g. "000/0.jpg"
        new_file_name = f"images/{idx // 1000:03d}/{idx}.jpg"

        new_annotation = annotation_image["annotation"].copy()
        new_annotation["image_id"] = new_image_id
        new_test_metadata["annotations"].append(new_annotation)

        new_image = annotation_image["image"].copy()
        new_image["id"] = new_image_id
        new_image["file_name"] = new_file_name
        new_test_metadata["images"].append(new_image)

        # Copy file from raw to public
        if not dev_mode or idx < dev_count:  # if dev_mode, only copy the first dev_count images
            src_path = raw / "nybg2020/train" / annotation_image["image"]["file_name"]
            dst_path = public / "nybg2020/test" / new_file_name
            dst_path.parent.mkdir(exist_ok=True, parents=True)
            shutil.copyfile(src=src_path, dst=dst_path)

    # Save new test metadata
    with open(public / "nybg2020/test/metadata.json", "w") as f:
        # The public test data, of course, doesn't have annotations
        public_new_test = new_test_metadata.copy()
        del public_new_test["annotations"]
        assert public_new_test.keys() == {
            "images",
            "info",
            "licenses",
        }, f"Public test metadata keys should be 'images', 'info', 'licenses', but found {public_new_test.keys()}"
        json.dump(public_new_test, f, indent=4, sort_keys=True)

    if not dev_mode:
        assert len(list((public / "nybg2020/test/images").glob("**/*.jpg"))) == len(
            new_test_metadata["images"]
        ), f"Mismatching number of images in test_images, got {len(list((public / 'nybg2020/test/images').glob('**/*.jpg')))}"
        assert len(new_test_metadata["annotations"]) == len(
            new_test_metadata["images"]
        ), f"Mismatching number of annotations in test_metadata, got {len(new_test_metadata['annotations'])}"
        assert len(new_train_metadata["annotations"]) + len(
            new_test_metadata["annotations"]
        ) == len(old_train_metadata["annotations"]), (
            f"Expected {len(old_train_metadata['annotations'])} annotations in total, but found"
            f"{len(new_train_metadata['annotations'])} in train and {len(new_test_metadata['annotations'])} in test"
        )

    # Save private test answers
    answers_rows = []
    for image, annotation in zip(new_test_metadata["images"], new_test_metadata["annotations"]):
        assert (
            image["id"] == annotation["image_id"]
        ), f"Mismatching image_id in image and annotation: {image['id']} vs {annotation['image_id']}"
        answers_rows.append(
            {
                "Id": image["id"],
                "Predicted": annotation["category_id"],
            }
        )
    answers_df = pd.DataFrame(answers_rows)
    answers_df.to_csv(private / "answers.csv", index=False)

    # Create new sample submission that matches raw/sample_submission.csv, but for the new test set
    sample_rows = []
    for image in new_test_metadata["images"]:
        sample_rows.append(
            {
                "Id": image["id"],
                "Predicted": 0,
            }
        )
    sample_df = pd.DataFrame(sample_rows)
    sample_df.to_csv(public / "sample_submission.csv", index=False)

    assert len(answers_df) == len(
        new_test_metadata["images"]
    ), f"Expected {len(new_test_metadata['images'])} rows in answers, but found {len(answers_df)}"
    assert len(sample_df) == len(
        answers_df
    ), f"Expected {len(answers_df)} rows in sample submission, but found {len(sample_df)}"
    assert answers_df["Id"].equals(
        sample_df["Id"]
    ), "Mismatched 'Id' columns between answers and sample submission"



import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import List

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import extract

CLASSES = [
    "yes",
    "no",
    "up",
    "down",
    "left",
    "right",
    "on",
    "off",
    "stop",
    "go",
    "unknown",
    "silence",
]


@dataclass(frozen=True)
class AudioFile:
    label: str
    path: Path


def _create_and_populate_split(
    files_to_split: List[AudioFile],
    public_dir: Path,
    private_dir: Path,
    test_size: float,
    random_state: int,
) -> List[AudioFile]:
    """
    Splits a list of audio files into train and test sets, and populates the
    corresponding public and private directories.

    Args:
        files_to_split: A list of AudioFile objects to be split.
        public_dir: The public directory to save training data and test stubs.
        private_dir: The private directory to save test ground truth.
        test_size: The proportion of the dataset to allocate to the test split.
        random_state: The seed used by the random number generator.

    Returns:
        A list of AudioFile objects that were assigned to the training set.
    """
    train_files, test_files = train_test_split(
        files_to_split, test_size=test_size, random_state=random_state
    )

    # Make necessary directories
    public_dir.mkdir(parents=True, exist_ok=True)
    private_dir.mkdir(parents=True, exist_ok=True)

    labels = list(
        dict.fromkeys([file.label for file in train_files])
    )  # Gets unique elements deterministically

    for label in labels:
        (public_dir / "train" / "audio" / label).mkdir(parents=True, exist_ok=True)

    (public_dir / "test" / "audio").mkdir(parents=True, exist_ok=True)

    # Copy over train and test files
    desc_suffix = public_dir.name
    for file in tqdm(train_files, desc=f"Copying train files to {desc_suffix}"):
        shutil.copyfile(
            src=file.path,
            dst=public_dir / "train" / "audio" / file.label / file.path.name,
        )

    test_records = []

    for idx, file in enumerate(tqdm(test_files, desc=f"Copying test files to {desc_suffix}")):
        # Rename files, since training audio files across labels aren't necessarily unique.
        new_id = str(idx).zfill(8)
        new_name = f"clip_{new_id}.wav"
        test_records.append({"fname": new_name, "label": file.label})

        shutil.copyfile(
            src=file.path,
            dst=public_dir / "test" / "audio" / new_name,
        )

    test = pd.DataFrame.from_records(test_records)
    test.to_csv(private_dir / "test.csv", index=False)

    test_without_labels = test.drop(columns=["label"])
    sample_submission = test_without_labels.copy()
    sample_submission["label"] = "silence"
    sample_submission.to_csv(public_dir / "sample_submission.csv", index=False)

    # Sanity checks
    test_audio_files = list((public_dir / "test" / "audio").glob("*.wav"))
    num_test_files = len(test_audio_files)
    num_submission_entries = len(sample_submission)
    assert num_test_files == num_submission_entries, (
        f"The number of test audio files ({num_test_files}) does not match the number of entries "
        f"in sample_submission.csv ({num_submission_entries}). Please check the file copying process."
    )

    return train_files


def prepare(raw: Path, public: Path, private: Path):
    # extract only what we need
    extract(raw / "train.7z", raw)

    # Create train, test from train split
    audio_dir = raw / "train" / "audio"
    all_audio_files = sorted(
        [AudioFile(fpath.parent.name, fpath) for fpath in audio_dir.rglob("*.wav")],
        key=lambda x: f"{x.label}_{x.path.name}",
    )

    # --- First Split: Create the original train/test sets ---
    # This call populates the original `public` and `private` directories.
    # It returns the files assigned to the training set for the next split.
    train_files_for_val_split = _create_and_populate_split(
        files_to_split=all_audio_files,
        public_dir=public,
        private_dir=private,
        test_size=0.1,
        random_state=0,
    )

    # --- Second Split: Create the new train_val/test_val sets ---
    # Define new paths for the validation split, parallel to the original ones.
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # To get a validation test set of the same size as the original test set,
    # we take a 1/9 slice of the remaining 90% of the data (0.1 / 0.9 = 1/9).
    _create_and_populate_split(
        files_to_split=train_files_for_val_split,
        public_dir=public_val,
        private_dir=private_val,
        test_size=1 / 9,
        random_state=0, # Use the same random_state for deterministic behavior
    )


import numpy as np
import pandas as pd
from pandas import DataFrame

from mlebench.grade_helpers import InvalidSubmissionError


def prepare_for_metric(submission: DataFrame, answers: DataFrame) -> dict:

    assert "Patient_Week" in answers.columns, "Answers DataFrame must have a 'Patient_Week' column."
    assert "FVC" in answers.columns, "Answers DataFrame must have a 'FVC' column."
    assert "Patient" in answers.columns, "Answers DataFrame must have a 'Patient' column."
    if "Patient_Week" not in submission.columns:
        raise InvalidSubmissionError("Submission DataFrame must have a 'Patient_Week' column.")
    if "FVC" not in submission.columns:
        raise InvalidSubmissionError("Submission DataFrame must have a 'FVC' column.")
    if "Confidence" not in submission.columns:
        raise InvalidSubmissionError("Submission DataFrame must have a 'Confidence' column.")
    for pw in submission["Patient_Week"]:
        if pw not in answers["Patient_Week"].values:
            raise InvalidSubmissionError(
                f"Patient_Week {pw} in submission does not exist in answers"
            )
    if not pd.api.types.is_numeric_dtype(submission["FVC"]):
        raise InvalidSubmissionError("FVC column in submission must be numeric.")
    if not pd.api.types.is_numeric_dtype(submission["Confidence"]):
        raise InvalidSubmissionError("Confidence column in submission must be numeric.")

    # fillna with 0 for the confidence column
    submission["Confidence"] = submission["Confidence"].fillna(0)

    # We should only take the 3 latest Patient_Week records for each patient
    answers = answers.dropna(subset=["FVC"])  # Drop the dummy data
    answers["Week_Number"] = answers["Patient_Week"].apply(lambda x: int(x.split("_")[-1]))
    latest_weeks = answers.sort_values("Week_Number").groupby("Patient").tail(3)
    answers = latest_weeks.drop(columns=["Week_Number"])
    # Make submission match; we only grade the prediction for the 3 latest weeks
    submission = submission[submission["Patient_Week"].isin(answers["Patient_Week"])]

    submission = submission.sort_values(by="Patient_Week")
    answers = answers.sort_values(by="Patient_Week")

    fvc_true = answers.loc[answers["Patient_Week"].isin(submission["Patient_Week"]), "FVC"].values
    fvc_pred = submission.loc[
        submission["Patient_Week"].isin(answers["Patient_Week"]), "FVC"
    ].values
    confidence = submission.loc[
        submission["Patient_Week"].isin(answers["Patient_Week"]), "Confidence"
    ].values

    return {"fvc_true": fvc_true, "fvc_pred": fvc_pred, "confidence": confidence}


def laplace_log_likelihood(
    actual_fvc: np.ndarray, predicted_fvc: np.ndarray, confidence: np.ndarray, return_values=False
) -> float:
    """
    Calculates the modified Laplace Log Likelihood score for osic-pulmonary-fibrosis-progression
    See https://www.kaggle.com/code/rohanrao/osic-understanding-laplace-log-likelihood for details on this implementation
    """
    sd_clipped = np.maximum(confidence, 70)
    delta = np.minimum(np.abs(actual_fvc - predicted_fvc), 1000)
    metric = -np.sqrt(2) * delta / sd_clipped - np.log(np.sqrt(2) * sd_clipped)

    if return_values:
        return metric
    else:
        return np.mean(metric)


def grade(submission: DataFrame, answers: DataFrame) -> float:
    fvc_true, fvc_pred, confidence = prepare_for_metric(submission, answers).values()

    return laplace_log_likelihood(fvc_true, fvc_pred, confidence)



import numpy as np


def average_precision_at_k(actual: set, predicted: list, k: int):
    """
    Computes the average precision at k (AP@k).

    This function computes the average precision at k between a predicted ranking and a ground truth
    set.

    Args:
        actual : A set of elements that are to be predicted (order doesn't matter)
        predicted : A list of predicted elements (order does matter, most relevant go first)
        k : The maximum number of predicted elements

    Adapted from: https://github.com/benhamner/Metrics/blob/master/Python/ml_metrics/average_precision.py
    """
    if len(predicted) > k:
        predicted = predicted[:k]

    score = 0.0
    num_hits = 0.0

    for i, p in enumerate(predicted):
        if p in actual and p not in predicted[:i]:
            num_hits += 1.0
            score += num_hits / (i + 1.0)

    if not actual:
        return 0.0

    return score / min(len(actual), k)


def mean_average_precision_at_k(actual: list[set], predicted: list[list], k: int):
    """
    Computes the MAP@k

    Args:
        actual : a list of sets of the elements that are to be predicted (order doesn't matter)
        predicted : a list of lists of predicted elements (order does matter, most relevant go first)
        k : The maximum number of predicted elements
    """
    return np.mean([average_precision_at_k(a, p, k) for a, p, in zip(actual, predicted)])


def dice_coefficient(
    predicted_mask: np.ndarray, true_mask: np.ndarray, both_empty_value: float = np.nan
) -> float:
    """
    Computes the Dice coefficient between two binary masks (can be multi-dimensional)

    Args:
        predicted_mask: A binary numpy array indicating where the segmentation is predicted
        true_mask: A binary numpy array indicating where the segmentation is
        both_empty_value: The value to return when both masks are empty
    """
    assert (
        predicted_mask.shape == true_mask.shape
    ), f"Predicted mask shape {predicted_mask.shape} does not match true mask shape {true_mask.shape}"
    # Check if both masks are empty
    if np.sum(predicted_mask) == 0 and np.sum(true_mask) == 0:
        return both_empty_value

    # Calculate intersection and union
    intersection = np.sum(predicted_mask * true_mask)
    union = np.sum(predicted_mask) + np.sum(true_mask)

    if union == 0:
        return both_empty_value

    # Calculate Dice coefficient
    dice_coeff = 2 * intersection / union

    return dice_coeff



from pandas import DataFrame
from sklearn.metrics import accuracy_score

from mlebench.grade_helpers import InvalidSubmissionError


def grade(submission: DataFrame, answers: DataFrame) -> float:
    expected_columns = set(["image_id", "label"])

    assert set(answers.columns).issuperset(
        expected_columns
    ), f"Expected columns {expected_columns}, but got {set(answers.columns)}"

    assert not answers.isnull().values.any(), "Answers contain NaN values"

    if not set(submission.columns).issuperset(expected_columns):
        raise InvalidSubmissionError(
            f"Expected columns {expected_columns}, but got {set(submission.columns)}"
        )

    if len(submission) != len(answers):
        raise InvalidSubmissionError(f"Expected {len(answers)} rows, but got {len(submission)}")

    sorted_submission = submission.sort_values(by="image_id").reset_index(drop=True)
    sorted_answers = answers.sort_values(by="image_id").reset_index(drop=True)

    if (sorted_submission["image_id"].values != sorted_answers["image_id"].values).any():
        raise InvalidSubmissionError("Invalid image IDs in the submission!")

    y_pred = sorted_submission["label"].fillna("").tolist()
    y_true = sorted_answers["label"]

    score = accuracy_score(y_true=y_true, y_pred=y_pred)

    return score



import shutil
from pathlib import Path

from pandas import DataFrame
from sklearn.model_selection import train_test_split

from mlebench.competitions.utils import df_to_one_hot
from mlebench.utils import extract, read_csv

from .classes import CLASSES


def _create_split_and_save(
    source_df: DataFrame,
    image_source_dir: Path,
    public_dir: Path,
    private_dir: Path,
    test_size: float,
    random_state: int,
) -> DataFrame:
    """Helper function to perform a split, copy images, and save all artifacts."""
    # Create train, test from the source dataframe
    new_train, new_test = train_test_split(
        source_df, test_size=test_size, random_state=random_state
    )
    new_test_without_labels = new_test.drop(columns=["species"])

    # match the format of the sample submission
    new_test = new_test[["id", "species"]]
    new_test = df_to_one_hot(new_test, "id", "species", classes=CLASSES)

    # Create destination directories
    (public_dir / "images").mkdir(parents=True, exist_ok=True)
    (private_dir / "images").mkdir(parents=True, exist_ok=True)

    # Copy images for the new training set
    for file_id in new_train["id"]:
        shutil.copyfile(
            src=image_source_dir / f"{file_id}.jpg",
            dst=public_dir / "images" / f"{file_id}.jpg",
        )

    # Copy images for the new test set
    for file_id in new_test_without_labels["id"]:
        shutil.copyfile(
            src=image_source_dir / f"{file_id}.jpg",
            dst=public_dir / "images" / f"{file_id}.jpg",
        )

    # Check integrity of the files copied
    assert len(new_test_without_labels) == len(
        new_test
    ), "Public and Private tests should have equal length"
    assert len(list(public_dir.glob("images/*.jpg"))) == len(new_train) + len(
        new_test_without_labels
    ), "Public images should have the same number of images as the sum of train and test"

    # Create a sample submission file
    submission_df = new_test.copy()
    submission_df[CLASSES] = 1 / len(CLASSES)

    # Save all dataframes to their respective files
    new_train.to_csv(public_dir / "train.csv", index=False)
    new_test.to_csv(private_dir / "test.csv", index=False)
    new_test_without_labels.to_csv(public_dir / "test.csv", index=False)
    submission_df.to_csv(public_dir / "sample_submission.csv", index=False)

    return new_train


def prepare(raw: Path, public: Path, private: Path):
    """
    Splits the data in raw into public and private datasets with appropriate test/train splits.
    Also creates a secondary validation split in public_val/private_val directories.
    """
    # extract only what we need
    extract(raw / "train.csv.zip", raw)
    extract(raw / "images.zip", raw)

    # Load the full raw training data
    full_train_df = read_csv(raw / "train.csv")
    image_source_dir = raw / "images"

    # --- Stage 1: Create the original train/test split ---
    # This block generates the primary competition data in `public/` and `private/`.
    # The outputs of this stage must remain unchanged from the original script.
    train_df_after_first_split = _create_split_and_save(
        source_df=full_train_df,
        image_source_dir=image_source_dir,
        public_dir=public,
        private_dir=private,
        test_size=0.1,
        random_state=0,
    )

    # --- Stage 2: Create the new train/validation split ---
    # This block takes the training set from Stage 1 and splits it again
    # to create a new, smaller training set and a validation set.
    # The outputs are saved to new `public_val/` and `private_val/` directories.

    # Define paths for the new validation split
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # Calculate test_size to make the new validation set have the same number
    # of samples as the original test set (0.1 / (1-0.1) = 0.1 / 0.9)
    validation_test_size = 0.1 / 0.9

    _create_split_and_save(
        source_df=train_df_after_first_split,
        image_source_dir=image_source_dir,
        public_dir=public_val,
        private_dir=private_val,
        test_size=validation_test_size,
        random_state=0,  # Use the same random state for consistency
    )


from pathlib import Path

import pandas as pd


def prepare(raw: Path, public: Path, private: Path) -> None:
    old_train = pd.read_json(raw / "train.json", lines=True)
    old_test = pd.read_json(raw / "test.json", lines=True)
    old_sample_submission = pd.read_csv(raw / "sample_submission.csv")

    to_predict = ["reactivity", "deg_Mg_pH10", "deg_pH10", "deg_Mg_50C", "deg_50C"]
    test_size = 0.1
    n_test_samples = int(len(old_train) * test_size)

    # only put samples that pass the SN filter in the test set, as per comp data desc
    old_train["test"] = False
    test_indices = (
        old_train[old_train["SN_filter"] > 0].sample(n=n_test_samples, random_state=0).index
    )
    old_train.loc[test_indices, "test"] = True

    new_train = old_train[~old_train["test"]].copy().drop(columns=["test"], inplace=False)
    new_test = old_train[old_train["test"]].copy().drop(columns=["test"], inplace=False)
    old_train = old_train.drop(columns=["test"], inplace=False)

    # Create `test.csv` by exploding each list in the `reactivity` and `deg_*` columns, analogous
    # to `pd.explode`. Only the first `seq_scored` items are scored out of a possible `seq_length`
    # items. For each row, we keep track of whether it's scored or not with the `keep` column.
    records = []

    for _, row in new_test.iterrows():
        n = row["seq_scored"]

        assert len(row["reactivity"]) == n
        assert len(row["deg_Mg_pH10"]) == n
        assert len(row["deg_pH10"]) == n
        assert len(row["deg_Mg_50C"]) == n
        assert len(row["deg_50C"]) == n

        for j in range(n):
            records.append(
                {
                    "id_seqpos": f"{row['id']}_{j}",
                    "reactivity": row["reactivity"][j],
                    "deg_Mg_pH10": row["deg_Mg_pH10"][j],
                    "deg_pH10": row["deg_pH10"][j],
                    "deg_Mg_50C": row["deg_Mg_50C"][j],
                    "deg_50C": row["deg_50C"][j],
                    "keep": True,
                }
            )

        k = row["seq_length"]

        assert n < k

        for j in range(n, k):
            records.append(
                {
                    "id_seqpos": f"{row['id']}_{j}",
                    "reactivity": 0.0,
                    "deg_Mg_pH10": 0.0,
                    "deg_pH10": 0.0,
                    "deg_Mg_50C": 0.0,
                    "deg_50C": 0.0,
                    "keep": False,
                }
            )

    # Write `answers.csv`
    answers = pd.DataFrame(records)
    answers.to_csv(private / "test.csv", index=False, float_format="%.10f")

    # Write `train.json`
    new_train["index"] = range(len(new_train))
    new_train.to_json(public / "train.json", orient="records", lines=True)

    # Write `test.json`
    new_test_without_labels = new_test[old_test.columns].copy()
    new_test_without_labels["index"] = range(len(new_test_without_labels))
    new_test_without_labels.to_json(public / "test.json", orient="records", lines=True)

    # Write `sample_submission.csv`
    new_sample_submission = answers[["id_seqpos"] + to_predict].copy()
    new_sample_submission.loc[:, to_predict] = 0.0
    new_sample_submission.to_csv(
        public / "sample_submission.csv", index=False, float_format="%.10f"
    )

    # Sanity checks
    assert set(new_train.columns) == set(old_train.columns), (
        f"Expected the columns of the new train to be the same as the old train, but got "
        f"{set(new_train.columns)} instead of {set(old_train.columns)}."
    )

    assert set(new_test_without_labels.columns) == set(old_test.columns), (
        f"Expected the columns of the new test to be the same as the old test, but got "
        f"{set(new_test_without_labels.columns)} instead of {set(old_test.columns)}."
    )

    assert set(to_predict).intersection(set(new_test_without_labels.columns)) == set(), (
        f"Expected the columns to predict aren't included in the new test, but got "
        f"{set(to_predict) ^ set(new_test_without_labels.columns)} instead of the empty set."
    )

    assert set(new_sample_submission.columns) == set(old_sample_submission.columns), (
        f"Expected the columns of the new sample submission to be the same as the old sample "
        f"submission, but got {set(new_sample_submission.columns)} instead of "
        f"{set(old_sample_submission.columns)}."
    )

    assert len(answers) == len(new_sample_submission), (
        f"Expected the answers to have the same length as the new sample submission, but got "
        f"{len(answers)} instead of {len(new_sample_submission)}."
    )

    # we can use [0] because all sequences have the same length
    assert len(new_sample_submission) == (
        len(new_test_without_labels) * new_test_without_labels["seq_length"].iloc[0]
    ), (
        "Expected new_sample_submission length to be equal to max seq_length * len(new_test)."
        f"Got {len(new_sample_submission)} instead of {len(new_test_without_labels) * new_test_without_labels['seq_length']}."
    )

    assert len(new_train) + len(new_test) == len(old_train), (
        f"Expected the length of the new train set plus the length of the new test set to be "
        f"equal to the length of the old train set, but got {len(new_train) + len(new_test)} "
        f"instead of {len(old_train)}."
    )



import shutil
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split
from tqdm import tqdm


def _create_split(
    df_to_split: pd.DataFrame,
    image_source_dir: Path,
    test_split_size: float,
    public_dest: Path,
    private_dest: Path,
    random_state: int,
) -> pd.DataFrame:
    """
    Splits a dataframe of labels, copies corresponding images, and saves the results.

    Args:
        df_to_split: DataFrame containing image IDs and labels to be split.
        image_source_dir: Path to the directory containing the source images.
        test_split_size: The proportion of the dataset to allocate to the test split.
        public_dest: The destination directory for public-facing files (train set, test images, sample submission).
        private_dest: The destination directory for private files (test set answers).
        random_state: The seed used by the random number generator for the split.

    Returns:
        The training portion of the split as a pandas DataFrame.
    """
    # Create destination directories
    public_dest.mkdir(exist_ok=True, parents=True)
    private_dest.mkdir(exist_ok=True, parents=True)
    (public_dest / "train").mkdir(exist_ok=True)
    (public_dest / "test").mkdir(exist_ok=True)

    # Perform the split
    train_ids, test_ids = train_test_split(
        df_to_split["id"], test_size=test_split_size, random_state=random_state
    )
    train_df = df_to_split[df_to_split["id"].isin(train_ids)]
    test_df = df_to_split[df_to_split["id"].isin(test_ids)]

    assert set(train_df["id"]).isdisjoint(
        set(test_df["id"])
    ), "Train should not contain id's of test images"
    assert len(train_ids) + len(test_ids) == len(
        df_to_split
    ), "The combined length of train_ids and test_ids should equal the length of the source df"

    # Copy over image files
    for file_id in tqdm(train_ids, desc=f"Copying train images to {public_dest}"):
        shutil.copyfile(
            src=image_source_dir / f"{file_id}.tif",
            dst=public_dest / "train" / f"{file_id}.tif",
        )
    for file_id in tqdm(test_ids, desc=f"Copying test images to {public_dest}"):
        shutil.copyfile(
            src=image_source_dir / f"{file_id}.tif",
            dst=public_dest / "test" / f"{file_id}.tif",
        )

    # Create and save label/submission files
    sample_submission = test_df.copy()
    sample_submission["label"] = 0

    train_df.to_csv(public_dest / "train_labels.csv", index=False)
    test_df.to_csv(private_dest / "answers.csv", index=False)
    sample_submission.to_csv(public_dest / "sample_submission.csv", index=False)

    # Check integrity of files copied
    assert len(list(public_dest.glob("train/*.tif"))) == len(
        train_ids
    ), "Number of train images should be equal to the number of unique id's in the train set"
    assert len(list(public_dest.glob("test/*.tif"))) == len(
        test_ids
    ), "Number of test images should be equal to the number of unique id's in the test set"

    return train_df


def prepare(raw: Path, public: Path, private: Path):
    """
    Prepares the data for the competition by performing two splits.
    1. A main split of the raw data into a primary train/test set for the competition.
       Outputs are saved to `public/` and `private/`.
    2. A secondary split of the primary train set into a smaller train/validation set for user convenience.
       Outputs are saved to `public_val/` and `private_val/`, mirroring the main structure.
    """
    # Common setup
    all_train_labels = pd.read_csv(raw / "train_labels.csv")
    image_source_dir = raw / "train"
    RANDOM_STATE = 0

    # --- 1. Main Split: Create the primary train/test sets ---
    # This logic is identical to the original script to ensure outputs do not change.
    num_test_from_pool = len(list((raw / "test").glob("*.tif")))
    test_ratio_main = num_test_from_pool / (len(all_train_labels) + num_test_from_pool)

    # The returned `primary_train_df` is the larger portion of the first split.
    primary_train_df = _create_split(
        df_to_split=all_train_labels,
        image_source_dir=image_source_dir,
        test_split_size=test_ratio_main,
        public_dest=public,
        private_dest=private,
        random_state=RANDOM_STATE,
    )

    # --- 2. Validation Split: Create a secondary train/validation set ---
    # Define new output paths for the validation split
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # To get the same number of test samples as the main split, we must adjust the test_size
    # for this second split, which operates on a smaller dataset (the primary_train_df).
    if (1 - test_ratio_main) <= 0:
        # Avoid division by zero in the unlikely case the main split uses all data for test
        test_ratio_val = 0
    else:
        test_ratio_val = test_ratio_main / (1 - test_ratio_main)

    # Perform the second split on the primary training data
    _create_split(
        df_to_split=primary_train_df,
        image_source_dir=image_source_dir,
        test_split_size=test_ratio_val,
        public_dest=public_val,
        private_dest=private_val,
        random_state=RANDOM_STATE,
    )


import shutil
from pathlib import Path

from sklearn.model_selection import train_test_split
from tqdm import tqdm

from mlebench.utils import read_csv


def _process_split(
    train_df,
    test_df,
    public_path: Path,
    private_path: Path,
    raw_images_path: Path,
    dev: bool,
):
    """
    Helper function to process a single data split (train/test).

    It handles the creation of directories, writing of CSVs (train, answers,
    sample_submission), and copying of image files. This ensures a consistent
    output structure for any given split.
    """
    # Create output directories
    public_path.mkdir(exist_ok=True, parents=True)
    private_path.mkdir(exist_ok=True, parents=True)
    (public_path / "test").mkdir(exist_ok=True)
    (public_path / "train").mkdir(exist_ok=True)

    # The test_df is the ground truth for the test set
    answers = test_df

    # Create a sample submission from the answers dataframe
    sample_submission_df = answers.copy()
    sample_submission_df["attribute_ids"] = "0 1 2"

    # Checks
    assert len(answers) == len(
        sample_submission_df
    ), "Answers and sample submission should have the same length"
    assert (
        sample_submission_df.columns.tolist() == answers.columns.tolist()
    ), "Sample submission and answers should have the same columns"

    # Write CSVs
    train_df.to_csv(public_path / "train.csv", index=False)
    answers.to_csv(private_path / "answers.csv", index=False)
    sample_submission_df.to_csv(public_path / "sample_submission.csv", index=False)

    # If in dev mode, reduce the number of images to copy for faster execution
    if dev:
        train_df_to_copy = train_df.head(int(len(train_df) * 0.1))
        answers_to_copy = answers.head(int(len(answers) * 0.1))
    else:
        train_df_to_copy = train_df
        answers_to_copy = answers

    # Copy train images
    for file_id in tqdm(
        train_df_to_copy["id"], desc=f"Copying train images to {public_path.name}"
    ):
        shutil.copyfile(
            src=raw_images_path / f"{file_id}.png",
            dst=public_path / "train" / f"{file_id}.png",
        )

    # Copy test images
    for file_id in tqdm(
        answers_to_copy["id"], desc=f"Copying test images to {public_path.name}"
    ):
        shutil.copyfile(
            src=raw_images_path / f"{file_id}.png",
            dst=public_path / "test" / f"{file_id}.png",
        )

    # File copying checks
    assert len(list(public_path.glob("train/*.png"))) == len(
        train_df_to_copy
    ), f"Train images in {public_path.name} should match dataframe"
    assert len(list(public_path.glob("test/*.png"))) == len(
        answers_to_copy
    ), f"Test images in {public_path.name} should match dataframe"


def prepare(raw: Path, public: Path, private: Path):

    dev = False

    # --- 1. Original Split: Create main train and test sets ---

    # Create train, test from train split
    old_train = read_csv(raw / "train.csv")
    # 25958/(142119+ 25958) = 0.15 original test split
    main_train, main_test = train_test_split(old_train, test_size=0.15, random_state=0)

    # Checks
    assert len(main_test) + len(main_train) == len(
        old_train
    ), "New train and test should have the same length as the old train"
    assert (
        main_test.columns.tolist() == old_train.columns.tolist()
    ), "Answers and old_train should have the same columns"
    assert (
        main_train.columns.tolist() == old_train.columns.tolist()
    ), "New train and old train should have the same columns"
    assert set(main_train["id"]).isdisjoint(
        set(main_test["id"])
    ), "New train and answers should have no overlap of rows"

    # Process and save the original split to `public` and `private` directories
    _process_split(
        train_df=main_train,
        test_df=main_test,
        public_path=public,
        private_path=private,
        raw_images_path=raw / "train",
        dev=dev,
    )

    # Copy over metadata file for the original split
    shutil.copyfile(src=raw / "labels.csv", dst=public / "labels.csv")

    # --- 2. New Validation Split: Create validation train and test sets ---

    # Define paths for the new validation split, parallel to the original ones
    public_val = public.parent / "public_val"
    private_val = private.parent / "private_val"

    # The test_size for the second split is calculated to yield a validation set
    # of approximately the same size as the original test set.
    # Original test set size = 0.15 * N
    # Original train set size = 0.85 * N
    # We want: test_size_val * (0.85 * N) ~= 0.15 * N
    # So, test_size_val ~= 0.15 / 0.85
    test_size_for_val_split = 0.15 / (1 - 0.15)

    # Split the main_train set again to create a smaller training set and a validation set.
    # We use the same random_state to ensure this split is also deterministic.
    train_val, test_val = train_test_split(
        main_train, test_size=test_size_for_val_split, random_state=0
    )

    # Process and save the validation split to `public_val` and `private_val`.
    # The filenames and structure will be identical to the original split, but
    # will contain the new, smaller training set and the validation set.
    _process_split(
        train_df=train_val,
        test_df=test_val,
        public_path=public_val,
        private_path=private_val,
        raw_images_path=raw / "train",
        dev=dev,
    )

    # Copy over metadata file for the validation split
    shutil.copyfile(src=raw / "labels.csv", dst=public_val / "labels.csv")